@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2025 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
+
 #ifndef UNICODE
 #define UNICODE
 #endif
@@ -20,10 +42,6 @@
 #define WM_MOUSEWHEEL 0x020A
 #endif
 
-#ifndef PATH_MAX
-#define PATH_MAX 4096
-#endif
-
 #define MIN(x,y) ((x) < (y) ? (x) : (y))
 
 #define ID_ABOUT	0x1000
@@ -57,7 +75,7 @@ static char filename[PATH_MAX];
 #define SET_KEY(parent, name, value) \
 	RegSetValueExA(parent, name, 0, REG_SZ, (const BYTE *)(value), (DWORD)strlen(value) + 1)
 
-void install_app(char *argv0)
+static void install_app(char *argv0)
 {
 	char buf[512];
 	HKEY software, classes, mupdf, dotpdf, dotxps, dotepub, dotfb2;
@@ -197,7 +215,7 @@ int winquery(pdfapp_t *app, const char *query)
 	}
 }
 
-int winfilename(wchar_t *buf, int len)
+static int winfilename(wchar_t *buf, int len)
 {
 	OPENFILENAME ofn;
 	buf[0] = 0;
@@ -213,7 +231,7 @@ int winfilename(wchar_t *buf, int len)
 	return GetOpenFileNameW(&ofn);
 }
 
-int wingetcertpath(char *buf, int len)
+int wingetcertpath(pdfapp_t *app, char *buf, int len)
 {
 	wchar_t twbuf[PATH_MAX] = {0};
 	OPENFILENAME ofn;
@@ -232,7 +250,7 @@ int wingetcertpath(char *buf, int len)
 		int code = WideCharToMultiByte(CP_UTF8, 0, twbuf, -1, buf, MIN(PATH_MAX, len), NULL, NULL);
 		if (code == 0)
 		{
-			winerror(&gapp, "cannot convert filename to utf-8");
+			pdfapp_error(app, "cannot convert filename to utf-8");
 			return 0;
 		}
 
@@ -264,12 +282,12 @@ int wingetsavepath(pdfapp_t *app, char *buf, int len)
 		int code = WideCharToMultiByte(CP_UTF8, 0, twbuf, -1, buf, MIN(PATH_MAX, len), NULL, NULL);
 		if (code == 0)
 		{
-			winerror(&gapp, "cannot convert filename to utf-8");
+			pdfapp_error(app, "cannot convert filename to utf-8");
 			return 0;
 		}
 
 		wcscpy(wbuf, twbuf);
-		strcpy(filename, buf);
+		fz_strlcpy(filename, buf, sizeof filename);
 		return 1;
 	}
 	else
@@ -278,7 +296,7 @@ int wingetsavepath(pdfapp_t *app, char *buf, int len)
 	}
 }
 
-void winreplacefile(char *source, char *target)
+void winreplacefile(pdfapp_t *app, char *source, char *target)
 {
 	wchar_t wsource[PATH_MAX];
 	wchar_t wtarget[PATH_MAX];
@@ -286,14 +304,14 @@ void winreplacefile(char *source, char *target)
 	int sz = MultiByteToWideChar(CP_UTF8, 0, source, -1, wsource, PATH_MAX);
 	if (sz == 0)
 	{
-		winerror(&gapp, "cannot convert filename to Unicode");
+		pdfapp_error(app, "cannot convert filename to Unicode");
 		return;
 	}
 
 	sz = MultiByteToWideChar(CP_UTF8, 0, target, -1, wtarget, PATH_MAX);
 	if (sz == 0)
 	{
-		winerror(&gapp, "cannot convert filename to Unicode");
+		pdfapp_error(app, "cannot convert filename to Unicode");
 		return;
 	}
 
@@ -305,7 +323,7 @@ void winreplacefile(char *source, char *target)
 #endif
 }
 
-void wincopyfile(char *source, char *target)
+void wincopyfile(pdfapp_t *app, char *source, char *target)
 {
 	wchar_t wsource[PATH_MAX];
 	wchar_t wtarget[PATH_MAX];
@@ -313,21 +331,20 @@ void wincopyfile(char *source, char *target)
 	int sz = MultiByteToWideChar(CP_UTF8, 0, source, -1, wsource, PATH_MAX);
 	if (sz == 0)
 	{
-		winerror(&gapp, "cannot convert filename to Unicode");
+		pdfapp_error(app, "cannot convert filename to Unicode");
 		return;
 	}
 
 	sz = MultiByteToWideChar(CP_UTF8, 0, target, -1, wtarget, PATH_MAX);
 	if (sz == 0)
 	{
-		winerror(&gapp, "cannot convert filename to Unicode");
+		pdfapp_error(app, "cannot convert filename to Unicode");
 		return;
 	}
 
 	CopyFile(wsource, wtarget, FALSE);
 }
 
-static char pd_filename[256] = "The file is encrypted.";
 static char pd_password[256] = "";
 static wchar_t pd_passwordw[256] = {0};
 static char td_textinput[1024] = "";
@@ -338,13 +355,13 @@ static const char **cd_opts;
 static const char **cd_vals;
 static int pd_okay = 0;
 
-INT_PTR CALLBACK
+static INT_PTR CALLBACK
 dlogpassproc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 {
 	switch(message)
 	{
 	case WM_INITDIALOG:
-		SetDlgItemTextA(hwnd, 4, pd_filename);
+		SetDlgItemTextA(hwnd, 4, "The file is encrypted.");
 		return TRUE;
 	case WM_COMMAND:
 		switch(wParam)
@@ -365,7 +382,7 @@ dlogpassproc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 	return FALSE;
 }
 
-INT_PTR CALLBACK
+static INT_PTR CALLBACK
 dlogtextproc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 {
 	switch(message)
@@ -395,14 +412,14 @@ dlogtextproc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 			SetTextColor((HDC)wParam, RGB(255,0,0));
 			SetBkMode((HDC)wParam, TRANSPARENT);
 
-			return (INT)GetStockObject(NULL_BRUSH);
+			return (INT_PTR)GetStockObject(NULL_BRUSH);
 		}
 		break;
 	}
 	return FALSE;
 }
 
-INT_PTR CALLBACK
+static INT_PTR CALLBACK
 dlogchoiceproc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 {
 	HWND listbox;
@@ -452,7 +469,6 @@ dlogchoiceproc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 
 char *winpassword(pdfapp_t *app, char *filename)
 {
-	char buf[1024], *s;
 	int code;
 
 	if (password)
@@ -462,16 +478,9 @@ char *winpassword(pdfapp_t *app, char *filename)
 		return p;
 	}
 
-	strcpy(buf, filename);
-	s = buf;
-	if (strrchr(s, '\\')) s = strrchr(s, '\\') + 1;
-	if (strrchr(s, '/')) s = strrchr(s, '/') + 1;
-	if (strlen(s) > 32)
-		strcpy(s + 30, "...");
-	sprintf(pd_filename, "The file \"%s\" is encrypted.", s);
 	code = DialogBoxW(NULL, L"IDD_DLOGPASS", hwndframe, dlogpassproc);
 	if (code <= 0)
-		winerror(app, "cannot create password dialog");
+		pdfapp_error(app, "cannot create password dialog");
 	if (pd_okay)
 		return pd_password;
 	return NULL;
@@ -484,7 +493,7 @@ char *wintextinput(pdfapp_t *app, char *inittext, int retry)
 	fz_strlcpy(td_textinput, inittext ? inittext : "", sizeof td_textinput);
 	code = DialogBoxW(NULL, L"IDD_DLOGTEXT", hwndframe, dlogtextproc);
 	if (code <= 0)
-		winerror(app, "cannot create text input dialog");
+		pdfapp_error(app, "cannot create text input dialog");
 	if (pd_okay)
 		return td_textinput;
 	return NULL;
@@ -499,11 +508,11 @@ int winchoiceinput(pdfapp_t *app, int nopts, const char *opts[], int *nvals, con
 	cd_vals = vals;
 	code = DialogBoxW(NULL, L"IDD_DLOGLIST", hwndframe, dlogchoiceproc);
 	if (code <= 0)
-		winerror(app, "cannot create text input dialog");
+		pdfapp_error(app, "cannot create text input dialog");
 	return pd_okay;
 }
 
-INT_PTR CALLBACK
+static INT_PTR CALLBACK
 dloginfoproc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 {
 	char buf[256];
@@ -577,14 +586,14 @@ dloginfoproc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 	return FALSE;
 }
 
-void info()
+static void info()
 {
 	int code = DialogBoxW(NULL, L"IDD_DLOGINFO", hwndframe, dloginfoproc);
 	if (code <= 0)
-		winerror(&gapp, "cannot create info dialog");
+		pdfapp_error(&gapp, "cannot create info dialog");
 }
 
-INT_PTR CALLBACK
+static INT_PTR CALLBACK
 dlogaboutproc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 {
 	switch(message)
@@ -604,14 +613,14 @@ void winhelp(pdfapp_t*app)
 {
 	int code = DialogBoxW(NULL, L"IDD_DLOGABOUT", hwndframe, dlogaboutproc);
 	if (code <= 0)
-		winerror(&gapp, "cannot create help dialog");
+		pdfapp_error(&gapp, "cannot create help dialog");
 }
 
 /*
  * Main window
  */
 
-void winopen()
+static void winopen()
 {
 	WNDCLASS wc;
 	HMENU menu;
@@ -632,7 +641,7 @@ void winopen()
 	wc.lpszClassName = L"FrameWindow";
 	a = RegisterClassW(&wc);
 	if (!a)
-		winerror(&gapp, "cannot register frame window class");
+		pdfapp_error(&gapp, "cannot register frame window class");
 
 	/* Create and register window view class */
 	memset(&wc, 0, sizeof(wc));
@@ -648,7 +657,7 @@ void winopen()
 	wc.lpszClassName = L"ViewWindow";
 	a = RegisterClassW(&wc);
 	if (!a)
-		winerror(&gapp, "cannot register view window class");
+		pdfapp_error(&gapp, "cannot register view window class");
 
 	/* Get screen size */
 	SystemParametersInfo(SPI_GETWORKAREA, 0, &r, 0);
@@ -689,7 +698,7 @@ void winopen()
 	0, // program instance handle
 	0); // creation parameters
 	if (!hwndframe)
-		winerror(&gapp, "cannot create frame");
+		pdfapp_error(&gapp, "cannot create frame");
 
 	hwndview = CreateWindowW(L"ViewWindow", // window class name
 	NULL,
@@ -698,7 +707,7 @@ void winopen()
 	CW_USEDEFAULT, CW_USEDEFAULT,
 	hwndframe, 0, 0, 0);
 	if (!hwndview)
-		winerror(&gapp, "cannot create view");
+		pdfapp_error(&gapp, "cannot create view");
 
 	hdc = NULL;
 
@@ -760,7 +769,7 @@ void wintitle(pdfapp_t *app, char *title)
 	SetWindowTextW(hwndframe, wide);
 }
 
-void windrawrect(pdfapp_t *app, int x0, int y0, int x1, int y1)
+static void windrawrect(pdfapp_t *app, int x0, int y0, int x1, int y1)
 {
 	RECT r;
 	r.left = x0;
@@ -777,7 +786,7 @@ void windrawstring(pdfapp_t *app, int x, int y, char *s)
 	TextOutA(hdc, x, y - 12, s, (int)strlen(s));
 }
 
-void winblitsearch()
+static void winblitsearch()
 {
 	if (gapp.issearching)
 	{
@@ -788,7 +797,7 @@ void winblitsearch()
 	}
 }
 
-void winblit()
+static void winblit()
 {
 	int image_w = fz_pixmap_width(gapp.ctx, gapp.image);
 	int image_h = fz_pixmap_height(gapp.ctx, gapp.image);
@@ -817,7 +826,7 @@ void winblit()
 
 		if (image_n == 2)
 		{
-			int i = image_w * image_h;
+			size_t i = image_w * (size_t)image_h;
 			unsigned char *color = malloc(i*4);
 			unsigned char *s = samples;
 			unsigned char *d = color;
@@ -965,10 +974,11 @@ static void killtimer(pdfapp_t *app)
 	timer_pending = 0;
 }
 
-void handlekey(int c)
+static void handlekey(int c)
 {
 	int modifier = (GetAsyncKeyState(VK_SHIFT) < 0);
 	modifier |= ((GetAsyncKeyState(VK_CONTROL) < 0)<<2);
+	modifier |= ((GetAsyncKeyState(VK_MENU) < 0)<<3);
 
 	if (timer_pending)
 		killtimer(&gapp);
@@ -987,12 +997,11 @@ void handlekey(int c)
 	{
 		switch (c - 256)
 		{
-		case VK_F1: c = '?'; break;
 		case VK_ESCAPE: c = '\033'; break;
 		case VK_DOWN: c = 'j'; break;
 		case VK_UP: c = 'k'; break;
-		case VK_LEFT: c = 'b'; break;
-		case VK_RIGHT: c = ' '; break;
+		case VK_LEFT: c = 'h'; break;
+		case VK_RIGHT: c = 'l'; break;
 		case VK_PRIOR: c = ','; break;
 		case VK_NEXT: c = '.'; break;
 		}
@@ -1002,7 +1011,7 @@ void handlekey(int c)
 	winrepaint(&gapp);
 }
 
-void handlemouse(int x, int y, int btn, int state)
+static void handlemouse(int x, int y, int btn, int state)
 {
 	int modifier = (GetAsyncKeyState(VK_SHIFT) < 0);
 	modifier |= ((GetAsyncKeyState(VK_CONTROL) < 0)<<2);
@@ -1024,7 +1033,7 @@ void handlemouse(int x, int y, int btn, int state)
 	pdfapp_onmouse(&gapp, x, y, btn, modifier, state);
 }
 
-LRESULT CALLBACK
+static LRESULT CALLBACK
 frameproc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 {
 	switch(message)
@@ -1082,7 +1091,7 @@ frameproc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 	return DefWindowProc(hwnd, message, wParam, lParam);
 }
 
-LRESULT CALLBACK
+static LRESULT CALLBACK
 viewproc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 {
 	static int oldx = 0;
@@ -1187,6 +1196,8 @@ viewproc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 		switch (wParam)
 		{
 		case VK_F1:
+			winhelp(&gapp);
+			return 0;
 		case VK_LEFT:
 		case VK_UP:
 		case VK_PRIOR:
@@ -1200,6 +1211,18 @@ viewproc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 		}
 		return 1;
 
+	case WM_SYSKEYDOWN:
+		/* alt keys */
+		switch (wParam)
+		{
+		case VK_LEFT:
+		case VK_RIGHT:
+			handlekey(wParam + 256);
+			handlemouse(oldx, oldy, 0, 0);	/* update cursor */
+			return 0;
+		}
+		break;
+
 	/* unicode encoded chars, including escape, backspace etc... */
 	case WM_CHAR:
 		if (wParam < 256)
@@ -1271,6 +1294,7 @@ WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShow
 	MSG msg;
 	int code;
 	fz_context *ctx;
+	char *profile_name = NULL;
 	int kbps = 0;
 	int displayRes = get_system_dpi();
 	int c;
@@ -1281,11 +1305,16 @@ WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShow
 		MessageBoxA(NULL, "Cannot initialize MuPDF context.", "MuPDF: Error", MB_OK);
 		exit(1);
 	}
+
+	/* stderr goes nowhere. Get us a debug stream we have a chance
+	 * of seeing. */
+	fz_set_stddbg(ctx, fz_stdods(ctx));
+
 	pdfapp_init(ctx, &gapp);
 
 	argv = fz_argv_from_wargv(argc, wargv);
 
-	while ((c = fz_getopt(argc, argv, "Ip:r:A:C:W:H:S:U:Xb:")) != -1)
+	while ((c = fz_getopt(argc, argv, "Ip:r:A:C:W:H:S:U:Xb:c:")) != -1)
 	{
 		switch (c)
 		{
@@ -1298,6 +1327,7 @@ WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShow
 		case 'r': displayRes = fz_atoi(fz_optarg); break;
 		case 'I': gapp.invert = 1; break;
 		case 'A': fz_set_aa_level(ctx, fz_atoi(fz_optarg)); break;
+		case 'c': profile_name = fz_optarg; break;
 		case 'W': gapp.layout_w = fz_atoi(fz_optarg); break;
 		case 'H': gapp.layout_h = fz_atoi(fz_optarg); break;
 		case 'S': gapp.layout_em = fz_atoi(fz_optarg); break;
@@ -1317,7 +1347,7 @@ WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShow
 
 	if (fz_optind < argc)
 	{
-		strcpy(filename, argv[fz_optind++]);
+		fz_strlcpy(filename, argv[fz_optind++], sizeof filename);
 	}
 	else
 	{
@@ -1325,12 +1355,15 @@ WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShow
 			exit(0);
 		code = WideCharToMultiByte(CP_UTF8, 0, wbuf, -1, filename, sizeof filename, NULL, NULL);
 		if (code == 0)
-			winerror(&gapp, "cannot convert filename to utf-8");
+			pdfapp_error(&gapp, "cannot convert filename to utf-8");
 	}
 
 	if (fz_optind < argc)
 		gapp.pageno = atoi(argv[fz_optind++]);
 
+	if (profile_name)
+		pdfapp_load_profile(&gapp, profile_name);
+
 	if (kbps)
 		pdfapp_open_progressive(&gapp, filename, 0, kbps);
 	else
