@@ -318,43 +318,6 @@ def build_swig(
                     return ret;
                 }}
 
-
-                /* mupdfpy optimisation for copying pixmap. Copies first <n>
-                bytes of each pixel from <src> to <pm>. <pm> and <src> must
-                have same `.w` and `.h` */
-                void ll_fz_pixmap_copy( fz_pixmap* pm, const fz_pixmap* src, int n)
-                {{
-                    assert( pm->w == src->w);
-                    assert( pm->h == src->h);
-                    assert( n <= pm->n);
-                    assert( n <= src->n);
-
-                    if (pm->n == src->n)
-                    {{
-                        // identical samples
-                        assert( pm->stride == src->stride);
-                        memcpy( pm->samples, src->samples, pm->w * pm->h * pm->n);
-                    }}
-                    else
-                    {{
-                        for ( int y=0; y<pm->h; ++y)
-                        {{
-                            for ( int x=0; x<pm->w; ++x)
-                            {{
-                                memcpy(
-                                        pm->samples + pm->stride * y + pm->n * x,
-                                        src->samples + src->stride * y + src->n * x,
-                                        n
-                                        );
-                                if (pm->alpha)
-                                {{
-                                    src->samples[ src->stride * y + src->n * x] = 255;
-                                }}
-                            }}
-                        }}
-                    }}
-                }}
-
                 /* mupdfpy optimisation for copying raw data into pixmap. `samples` must
                 have enough data to fill the pixmap. */
                 void ll_fz_pixmap_copy_raw( fz_pixmap* pm, const void* samples)
@@ -616,29 +579,31 @@ def build_swig(
                     int n
                     )
             {{
+                #define PDF_NAME2(X) {rename.namespace_class('pdf_obj')}(PDF_NAME(X))
                 for ( int i=0; i<n; ++i)
                 {{
                     {rename.namespace_class('pdf_obj')} o = {rename.namespace_fn('pdf_array_get')}( old_annots, i);
                     if ({rename.namespace_fn('pdf_dict_gets')}( o, "IRT").m_internal)
                         continue;
-                    {rename.namespace_class('pdf_obj')} subtype = {rename.namespace_fn('pdf_dict_get')}( o, PDF_NAME(Subtype));
-                    if ( {rename.namespace_fn('pdf_name_eq')}( subtype, PDF_NAME(Link)))
+                    {rename.namespace_class('pdf_obj')} subtype = {rename.namespace_fn('pdf_dict_get')}( o, PDF_NAME2(Subtype));
+                    if ( {rename.namespace_fn('pdf_name_eq')}( subtype, PDF_NAME2(Link)))
                         continue;
-                    if ( {rename.namespace_fn('pdf_name_eq')}( subtype, PDF_NAME(Popup)))
+                    if ( {rename.namespace_fn('pdf_name_eq')}( subtype, PDF_NAME2(Popup)))
                         continue;
-                    if ( {rename.namespace_fn('pdf_name_eq')}( subtype, PDF_NAME(Widget)))
+                    if ( {rename.namespace_fn('pdf_name_eq')}( subtype, PDF_NAME2(Widget)))
                     {{
                         /* fixme: C++ API doesn't yet wrap fz_warn() - it
                         excludes all variadic fns. */
                         //mupdf::fz_warn( "skipping widget annotation");
                         continue;
                     }}
-                    {rename.namespace_fn('pdf_dict_del')}( o, PDF_NAME(Popup));
-                    {rename.namespace_fn('pdf_dict_del')}( o, PDF_NAME(P));
+                    {rename.namespace_fn('pdf_dict_del')}( o, PDF_NAME2(Popup));
+                    {rename.namespace_fn('pdf_dict_del')}( o, PDF_NAME2(P));
                     {rename.namespace_class('pdf_obj')} copy_o = {rename.namespace_fn('pdf_graft_mapped_object')}( graft_map, o);
                     {rename.namespace_class('pdf_obj')} annot = {rename.namespace_fn('pdf_new_indirect')}( doc_des, {rename.namespace_fn('pdf_to_num')}( copy_o), 0);
                     {rename.namespace_fn('pdf_array_push')}( new_annots, annot);
                 }}
+                #undef PDF_NAME2
             }}
             ''')
 
@@ -650,9 +615,11 @@ def build_swig(
     text += '%module(directors="1") mupdf\n'
 
     jlib.log(f'{build_dirs.Py_LIMITED_API=}')
+
+    text += f'%begin %{{\n'
+
     if build_dirs.Py_LIMITED_API:  # e.g. 0x03080000
         text += textwrap.dedent(f'''
-                %begin %{{
                 /* Use Python Stable ABI with earliest Python version that we
                 support. */
                 #define Py_LIMITED_API {build_dirs.Py_LIMITED_API}
@@ -666,9 +633,17 @@ def build_swig(
                 #ifndef PyBUF_WRITE
                     #define PyBUF_WRITE 0x200
                 #endif
-                %}}
                 ''')
 
+        text += textwrap.dedent(f'''
+                /* This seems to be necessary on some Windows machines with
+                Py_LIMITED_API, otherwise compilation can fail because free()
+                and malloc() are not declared. */
+                #include <stdlib.h>
+                ''')
+
+    text += f'%}}\n'
+
     # https://www.mono-project.com/docs/advanced/pinvoke/
     #
     # > Mono on all platforms currently uses UTF-8 encoding for all string
@@ -1000,6 +975,8 @@ def build_swig(
                 %template(map_string_int) map<std::string, int>;
                 %template(vectorq) vector<{rename.namespace_class("fz_quad")}>;
                 %template(vector_search_page2_hit) vector<fz_search_page2_hit>;
+                %template(vector_fz_font_ucs_gid) vector<fz_font_ucs_gid>;
+                %template(vector_fz_point) vector<fz_point>;
             }};
 
             // Make sure that operator++() gets converted to __next__().
