@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2022 Artifex Software, Inc.
+// Copyright (C) 2004-2024 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -28,10 +28,6 @@
 #include <float.h>
 #include <stdlib.h>
 
-#ifndef PATH_MAX
-#define PATH_MAX 4096
-#endif
-
 #ifdef _WIN32
 #include <windows.h> /* for MultiByteToWideChar etc. */
 #endif
@@ -742,6 +738,18 @@ int64_t fz_atoi64(const char *s)
 	return atoll(s);
 }
 
+size_t fz_atoz(const char *s)
+{
+	int64_t i;
+
+	if (s == NULL)
+		return 0;
+	i = atoll(s);
+	if (i < 0 || (int64_t)(size_t)i != i)
+		return 0;
+	return (size_t)i;
+}
+
 int fz_is_page_range(fz_context *ctx, const char *s)
 {
 	/* TODO: check the actual syntax... */
