@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2022 Artifex Software, Inc.
+// Copyright (C) 2004-2024 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -26,7 +26,9 @@
 #include <stdlib.h>
 #include <stdio.h>
 
+#if FZ_ENABLE_HTML_ENGINE
 #include <gumbo.h>
+#endif
 
 #define FZ_XML_MAX_DEPTH 4096
 
@@ -132,7 +134,9 @@ void fz_output_xml(fz_context *ctx, fz_output *out, fz_xml *item, int level)
 	/* Skip over the DOC object at the top. */
 	if (item->up == NULL)
 	{
-		fz_output_xml(ctx, out, item->down, level);
+		fz_xml *child;
+		for (child = fz_xml_down(item); child; child = child->u.node.next)
+			fz_output_xml(ctx, out, child, level + 1);
 		return;
 	}
 
@@ -1061,7 +1065,7 @@ fz_parse_xml(fz_context *ctx, fz_buffer *buf, int preserve_white)
 {
 	struct parser parser;
 	fz_xml *xml = NULL;
-	fz_xml root, *node;
+	fz_xml *root, *node;
 	char *p = NULL;
 	char *error;
 	int dofree = 0;
@@ -1084,9 +1088,8 @@ fz_parse_xml(fz_context *ctx, fz_buffer *buf, int preserve_white)
 		n = fz_buffer_storage(ctx, buf, &s);
 	}
 
-	memset(&root, 0, sizeof(root));
 	parser.pool = fz_new_pool(ctx);
-	parser.head = &root;
+	parser.head = root = fz_pool_alloc(ctx, parser.pool, offsetof(fz_xml, u.node.u.d.name) + 1);
 	parser.preserve_white = preserve_white;
 	parser.depth = 0;
 #ifdef FZ_XML_SEQ
@@ -1106,11 +1109,11 @@ fz_parse_xml(fz_context *ctx, fz_buffer *buf, int preserve_white)
 
 		xml = fz_pool_alloc(ctx, parser.pool, sizeof *xml);
 		xml->up = NULL;
-		xml->down = root.down;
+		xml->down = root->down;
 		xml->u.doc.refs = 1;
 		xml->u.doc.pool = parser.pool;
 
-		for (node = root.down; node; node = node->u.node.next)
+		for (node = root->down; node; node = node->u.node.next)
 			node->up = xml;
 	}
 	fz_always(ctx)
@@ -1127,6 +1130,7 @@ fz_parse_xml(fz_context *ctx, fz_buffer *buf, int preserve_white)
 	return xml;
 }
 
+#if FZ_ENABLE_HTML_ENGINE
 /*
 	Parse the contents of buffer into a tree of XML nodes, using the HTML5 syntax.
 
@@ -1201,10 +1205,12 @@ static void xml_from_gumbo(fz_context *ctx, struct parser *parser, GumboNode *no
 		break;
 	}
 }
+#endif
 
 fz_xml *
 fz_parse_xml_from_html5(fz_context *ctx, fz_buffer *buf)
 {
+#if FZ_ENABLE_HTML_ENGINE
 	struct parser parser;
 	fz_xml *xml = NULL;
 	fz_xml root, *node;
@@ -1292,6 +1298,9 @@ fz_parse_xml_from_html5(fz_context *ctx, fz_buffer *buf)
 	}
 
 	return xml;
+#else
+	fz_throw(ctx, FZ_ERROR_GENERIC, "HTML Engine not enabled in this build");
+#endif
 }
 
 fz_xml *fz_xml_find_dfs(fz_xml *item, const char *tag, const char *att, const char *match)
