@@ -1,11 +1,35 @@
+// Copyright (C) 2004-2024 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
+
 #include "mupdf/fitz.h"
 #include "mupdf/pdf.h"
 
 #include <string.h>
 
+static fz_colorspace *pdf_load_colorspace_imp(fz_context *ctx, pdf_obj *obj, pdf_cycle_list *cycle_up);
+
 /* ICCBased */
 static fz_colorspace *
-load_icc_based(fz_context *ctx, pdf_obj *dict, int allow_alt)
+load_icc_based(fz_context *ctx, pdf_obj *dict, int allow_alt, pdf_cycle_list *cycle_up)
 {
 	int n = pdf_dict_get_int(ctx, dict, PDF_NAME(N));
 	fz_colorspace *alt = NULL;
@@ -22,10 +46,12 @@ load_icc_based(fz_context *ctx, pdf_obj *dict, int allow_alt)
 		if (obj)
 		{
 			fz_try(ctx)
-				alt = pdf_load_colorspace(ctx, obj);
+				alt = pdf_load_colorspace_imp(ctx, obj, cycle_up);
 			fz_catch(ctx)
 			{
 				fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
+				fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
+				fz_report_error(ctx);
 				fz_warn(ctx, "ignoring broken ICC Alternate colorspace");
 			}
 		}
@@ -54,7 +80,12 @@ load_icc_based(fz_context *ctx, pdf_obj *dict, int allow_alt)
 			fz_drop_buffer(ctx, buf);
 		fz_catch(ctx)
 		{
-			fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
+			if (fz_caught(ctx) == FZ_ERROR_TRYLATER || fz_caught(ctx) == FZ_ERROR_SYSTEM)
+			{
+				fz_drop_colorspace(ctx, alt);
+				fz_rethrow(ctx);
+			}
+			fz_report_error(ctx);
 			fz_warn(ctx, "ignoring broken ICC profile");
 		}
 	}
@@ -89,7 +120,7 @@ devicen_drop(fz_context *ctx, void *tint)
 }
 
 static fz_colorspace *
-load_devicen(fz_context *ctx, pdf_obj *array, int is_devn)
+load_devicen(fz_context *ctx, pdf_obj *array, int is_devn, pdf_cycle_list *cycle_up)
 {
 	fz_colorspace *base = NULL;
 	fz_colorspace *cs = NULL;
@@ -99,6 +130,8 @@ load_devicen(fz_context *ctx, pdf_obj *array, int is_devn)
 	char name[100];
 	int i, n;
 
+	fz_var(cs);
+
 	if (pdf_is_array(ctx, nameobj))
 	{
 		n = pdf_array_len(ctx, nameobj);
@@ -112,7 +145,7 @@ load_devicen(fz_context *ctx, pdf_obj *array, int is_devn)
 		n = 1;
 	}
 
-	base = pdf_load_colorspace(ctx, baseobj);
+	base = pdf_load_colorspace_imp(ctx, baseobj, cycle_up);
 	fz_try(ctx)
 	{
 		if (is_devn)
@@ -136,7 +169,7 @@ load_devicen(fz_context *ctx, pdf_obj *array, int is_devn)
 		cs->u.separation.tint = pdf_load_function(ctx, tintobj, n, cs->u.separation.base->n);
 		if (pdf_is_array(ctx, nameobj))
 			for (i = 0; i < n; i++)
-				fz_colorspace_name_colorant(ctx, cs, i, pdf_to_name(ctx, pdf_array_get(ctx, nameobj, i)));
+				fz_colorspace_name_colorant(ctx, cs, i, pdf_array_get_name(ctx, nameobj, i));
 		else
 			fz_colorspace_name_colorant(ctx, cs, 0, pdf_to_name(ctx, nameobj));
 	}
@@ -162,7 +195,7 @@ pdf_is_tint_colorspace(fz_context *ctx, fz_colorspace *cs)
 /* Indexed */
 
 static fz_colorspace *
-load_indexed(fz_context *ctx, pdf_obj *array)
+load_indexed(fz_context *ctx, pdf_obj *array, pdf_cycle_list *cycle_up)
 {
 	pdf_obj *baseobj = pdf_array_get(ctx, array, 1);
 	pdf_obj *highobj = pdf_array_get(ctx, array, 2);
@@ -178,7 +211,7 @@ load_indexed(fz_context *ctx, pdf_obj *array)
 
 	fz_try(ctx)
 	{
-		base = pdf_load_colorspace(ctx, baseobj);
+		base = pdf_load_colorspace_imp(ctx, baseobj, cycle_up);
 
 		high = pdf_to_int(ctx, highobj);
 		high = fz_clampi(high, 0, 255);
@@ -331,10 +364,13 @@ pdf_load_cal_rgb(fz_context *ctx, pdf_obj *dict)
 /* Parse and create colorspace from PDF object */
 
 static fz_colorspace *
-pdf_load_colorspace_imp(fz_context *ctx, pdf_obj *obj)
+pdf_load_colorspace_imp(fz_context *ctx, pdf_obj *obj, pdf_cycle_list *cycle_up)
 {
-	if (pdf_obj_marked(ctx, obj))
-		fz_throw(ctx, FZ_ERROR_SYNTAX, "recursion in colorspace definition");
+	fz_colorspace *cs = NULL;
+	pdf_cycle_list cycle;
+
+	if (pdf_cycle(ctx, &cycle, cycle_up, obj))
+		fz_throw(ctx, FZ_ERROR_SYNTAX, "recursive colorspace");
 
 	if (pdf_is_name(ctx, obj))
 	{
@@ -375,58 +411,46 @@ pdf_load_colorspace_imp(fz_context *ctx, pdf_obj *obj)
 				return fz_keep_colorspace(ctx, fz_device_rgb(ctx));
 			else if (pdf_name_eq(ctx, name, PDF_NAME(DeviceCMYK)))
 				return fz_keep_colorspace(ctx, fz_device_cmyk(ctx));
-			else if (pdf_name_eq(ctx, name, PDF_NAME(CalGray)))
-				return pdf_load_cal_gray(ctx, pdf_array_get(ctx, obj, 1));
-			else if (pdf_name_eq(ctx, name, PDF_NAME(CalRGB)))
-				return pdf_load_cal_rgb(ctx, pdf_array_get(ctx, obj, 1));
 			else if (pdf_name_eq(ctx, name, PDF_NAME(CalCMYK)))
 				return fz_keep_colorspace(ctx, fz_device_cmyk(ctx));
 			else if (pdf_name_eq(ctx, name, PDF_NAME(Lab)))
 				return fz_keep_colorspace(ctx, fz_device_lab(ctx));
+
 			else
 			{
-				fz_colorspace *cs;
-				fz_try(ctx)
+				if ((cs = pdf_find_item(ctx, fz_drop_colorspace_imp, obj)) != NULL)
+					return cs;
+
+				if (pdf_name_eq(ctx, name, PDF_NAME(ICCBased)))
+					cs = load_icc_based(ctx, pdf_array_get(ctx, obj, 1), 1, &cycle);
+
+				else if (pdf_name_eq(ctx, name, PDF_NAME(CalGray)))
+					cs = pdf_load_cal_gray(ctx, pdf_array_get(ctx, obj, 1));
+				else if (pdf_name_eq(ctx, name, PDF_NAME(CalRGB)))
+					cs = pdf_load_cal_rgb(ctx, pdf_array_get(ctx, obj, 1));
+
+				else if (pdf_name_eq(ctx, name, PDF_NAME(Indexed)))
+					cs = load_indexed(ctx, obj, &cycle);
+				else if (pdf_name_eq(ctx, name, PDF_NAME(I)))
+					cs = load_indexed(ctx, obj, &cycle);
+
+				else if (pdf_name_eq(ctx, name, PDF_NAME(Separation)))
+					cs = load_devicen(ctx, obj, 0, &cycle);
+
+				else if (pdf_name_eq(ctx, name, PDF_NAME(DeviceN)))
+					cs = load_devicen(ctx, obj, 1, &cycle);
+
+				else if (pdf_name_eq(ctx, name, PDF_NAME(Pattern)))
 				{
-					if (pdf_mark_obj(ctx, obj))
-						fz_throw(ctx, FZ_ERROR_SYNTAX, "recursive colorspace");
-					if (pdf_name_eq(ctx, name, PDF_NAME(ICCBased)))
-						cs = load_icc_based(ctx, pdf_array_get(ctx, obj, 1), 1);
-
-					else if (pdf_name_eq(ctx, name, PDF_NAME(Indexed)))
-						cs = load_indexed(ctx, obj);
-					else if (pdf_name_eq(ctx, name, PDF_NAME(I)))
-						cs = load_indexed(ctx, obj);
-
-					else if (pdf_name_eq(ctx, name, PDF_NAME(Separation)))
-						cs = load_devicen(ctx, obj, 0);
-
-					else if (pdf_name_eq(ctx, name, PDF_NAME(DeviceN)))
-						cs = load_devicen(ctx, obj, 1);
-					else if (pdf_name_eq(ctx, name, PDF_NAME(Pattern)))
-					{
-						pdf_obj *pobj;
-
-						pobj = pdf_array_get(ctx, obj, 1);
-						if (!pobj)
-						{
-							cs = fz_keep_colorspace(ctx, fz_device_gray(ctx));
-							break;
-						}
-
-						cs = pdf_load_colorspace(ctx, pobj);
-					}
-					else
-						fz_throw(ctx, FZ_ERROR_SYNTAX, "unknown colorspace %s", pdf_to_name(ctx, name));
-				}
-				fz_always(ctx)
-				{
-					pdf_unmark_obj(ctx, obj);
-				}
-				fz_catch(ctx)
-				{
-					fz_rethrow(ctx);
+					pdf_obj *pobj = pdf_array_get(ctx, obj, 1);
+					if (!pobj)
+						return fz_keep_colorspace(ctx, fz_device_gray(ctx));
+					cs = pdf_load_colorspace_imp(ctx, pobj, &cycle);
 				}
+				else
+					fz_throw(ctx, FZ_ERROR_SYNTAX, "unknown colorspace %s", pdf_to_name(ctx, name));
+
+				pdf_store_item(ctx, obj, cs, 1000);
 				return cs;
 			}
 		}
@@ -439,7 +463,11 @@ pdf_load_colorspace_imp(fz_context *ctx, pdf_obj *obj)
 	 * handle it, so do our best. */
 	else if (pdf_is_dict(ctx, obj))
 	{
-		return load_icc_based(ctx, obj, 1);
+		if ((cs = pdf_find_item(ctx, fz_drop_colorspace_imp, obj)) != NULL)
+			return cs;
+		cs = load_icc_based(ctx, obj, 1, &cycle);
+		pdf_store_item(ctx, obj, cs, 1000);
+		return cs;
 	}
 
 	fz_throw(ctx, FZ_ERROR_SYNTAX, "could not parse color space (%d 0 R)", pdf_to_num(ctx, obj));
@@ -448,18 +476,7 @@ pdf_load_colorspace_imp(fz_context *ctx, pdf_obj *obj)
 fz_colorspace *
 pdf_load_colorspace(fz_context *ctx, pdf_obj *obj)
 {
-	fz_colorspace *cs;
-
-	if ((cs = pdf_find_item(ctx, fz_drop_colorspace_imp, obj)) != NULL)
-	{
-		return cs;
-	}
-
-	cs = pdf_load_colorspace_imp(ctx, obj);
-
-	pdf_store_item(ctx, obj, cs, 1000);
-
-	return cs;
+	return pdf_load_colorspace_imp(ctx, obj, NULL);
 }
 
 #if FZ_ENABLE_ICC
@@ -489,10 +506,12 @@ pdf_load_output_intent(fz_context *ctx, pdf_document *doc)
 	fz_var(cs);
 
 	fz_try(ctx)
-		cs = load_icc_based(ctx, dest_profile, 0);
+		cs = load_icc_based(ctx, dest_profile, 0, NULL);
 	fz_catch(ctx)
 	{
 		fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
+		fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
+		fz_report_error(ctx);
 		fz_warn(ctx, "Attempt to read Output Intent failed");
 	}
 	return cs;
