@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2022 Artifex Software, Inc.
+// Copyright (C) 2004-2024 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -17,8 +17,8 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 #include "mupdf/fitz.h"
 #include "mupdf/pdf.h"
@@ -171,13 +171,16 @@ const char *pdf_clean_font_name(const char *fontname)
 
 enum { UNKNOWN, TYPE1, TRUETYPE };
 
-static int ft_kind(FT_Face face)
+static int ft_kind(fz_context *ctx, FT_Face face)
 {
+	const char *kind;
+	fz_ft_lock(ctx);
 #ifdef FT_FONT_FORMATS_H
-	const char *kind = FT_Get_Font_Format(face);
+	kind = FT_Get_Font_Format(face);
 #else
-	const char *kind = FT_Get_X11_Font_Format(face);
+	kind = FT_Get_X11_Font_Format(face);
 #endif
+	fz_ft_unlock(ctx);
 	if (!strcmp(kind, "TrueType")) return TRUETYPE;
 	if (!strcmp(kind, "Type 1")) return TYPE1;
 	if (!strcmp(kind, "CFF")) return TYPE1;
@@ -260,7 +263,13 @@ int
 pdf_font_cid_to_gid(fz_context *ctx, pdf_font_desc *fontdesc, int cid)
 {
 	if (fontdesc->font->ft_face)
-		return ft_cid_to_gid(fontdesc, cid);
+	{
+		int gid;
+		fz_ft_lock(ctx);
+		gid = ft_cid_to_gid(fontdesc, cid);
+		fz_ft_unlock(ctx);
+		return gid;
+	}
 	return cid;
 }
 
@@ -429,21 +438,17 @@ pdf_load_substitute_cjk_font(fz_context *ctx, pdf_font_desc *fontdesc, const cha
 }
 
 static struct { int ros, serif; const char *name; } known_cjk_fonts[] = {
-	{ FZ_ADOBE_GB, 0, "SimFang" },
+	{ FZ_ADOBE_GB, 1, "SimFang" },
 	{ FZ_ADOBE_GB, 0, "SimHei" },
 	{ FZ_ADOBE_GB, 1, "SimKai" },
 	{ FZ_ADOBE_GB, 1, "SimLi" },
 	{ FZ_ADOBE_GB, 1, "SimSun" },
-	{ FZ_ADOBE_GB, 1, "NSimSun" },
 	{ FZ_ADOBE_GB, 1, "Song" },
 
 	{ FZ_ADOBE_CNS, 1, "MingLiU" },
-	{ FZ_ADOBE_CNS, 1, "PMingLiU" },
 
 	{ FZ_ADOBE_JAPAN, 0, "Gothic" },
-	{ FZ_ADOBE_JAPAN, 0, "PGothic" },
 	{ FZ_ADOBE_JAPAN, 1, "Mincho" },
-	{ FZ_ADOBE_JAPAN, 1, "PMincho" },
 
 	{ FZ_ADOBE_KOREA, 1, "Batang" },
 	{ FZ_ADOBE_KOREA, 0, "Gulim" },
@@ -452,10 +457,7 @@ static struct { int ros, serif; const char *name; } known_cjk_fonts[] = {
 
 static int match_font_name(const char *s, const char *ref)
 {
-	/* Skip "MS-" prefix if present. */
-	if (s[0] == 'M' && s[1] == 'S' && s[2] == '-')
-		return !strncmp(s+3, ref, strlen(ref));
-	return !strncmp(s, ref, strlen(ref));
+	return !!strstr(s, ref);
 }
 
 static void
@@ -528,7 +530,7 @@ pdf_extract_cff_subtable(fz_context *ctx, unsigned char *data, size_t size)
 	size_t i;
 
 	if (12 + num_tables * 16 > size)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "invalid TTF header");
+		fz_throw(ctx, FZ_ERROR_SYNTAX, "invalid TTF header");
 
 	for (i = 0; i < num_tables; ++i)
 	{
@@ -539,7 +541,7 @@ pdf_extract_cff_subtable(fz_context *ctx, unsigned char *data, size_t size)
 			uint64_t length = TTF_U32(record + 12);
 			uint64_t end = offset + length;
 			if (end > size)
-				fz_throw(ctx, FZ_ERROR_GENERIC, "invalid TTF subtable offset/length");
+				fz_throw(ctx, FZ_ERROR_SYNTAX, "invalid TTF subtable offset/length");
 			return fz_new_buffer_from_copied_data(ctx, data + offset, length);
 		}
 	}
@@ -600,6 +602,13 @@ pdf_drop_font(fz_context *ctx, pdf_font_desc *fontdesc)
 	fz_drop_storable(ctx, &fontdesc->storable);
 }
 
+static int
+pdf_font_is_droppable(fz_context *ctx, fz_storable *fontdesc)
+{
+	/* If we aren't holding the FT lock, then we can drop. */
+	return !fz_ft_lock_held(ctx);
+}
+
 static void
 pdf_drop_font_imp(fz_context *ctx, fz_storable *fontdesc_)
 {
@@ -622,7 +631,7 @@ pdf_new_font_desc(fz_context *ctx)
 	pdf_font_desc *fontdesc;
 
 	fontdesc = fz_malloc_struct(ctx, pdf_font_desc);
-	FZ_INIT_STORABLE(fontdesc, 1, pdf_drop_font_imp);
+	FZ_INIT_AWKWARD_STORABLE(fontdesc, 1, pdf_drop_font_imp, pdf_font_is_droppable);
 	fontdesc->size = sizeof(pdf_font_desc);
 
 	fontdesc->font = NULL;
@@ -685,7 +694,7 @@ select_type1_cmap(FT_Face face)
 }
 
 static FT_CharMap
-select_truetype_cmap(FT_Face face, int symbolic)
+select_truetype_cmap(fz_context *ctx, FT_Face face, int symbolic)
 {
 	int i;
 
@@ -697,21 +706,34 @@ select_truetype_cmap(FT_Face face, int symbolic)
 				return face->charmaps[i];
 	}
 
+	fz_ft_lock(ctx);
+
 	/* Then look for a Microsoft Unicode cmap */
 	for (i = 0; i < face->num_charmaps; i++)
 		if (face->charmaps[i]->platform_id == 3 && face->charmaps[i]->encoding_id == 1)
 			if (FT_Get_CMap_Format(face->charmaps[i]) != -1)
+			{
+				fz_ft_unlock(ctx);
 				return face->charmaps[i];
+			}
 
 	/* Finally look for an Apple MacRoman cmap */
 	for (i = 0; i < face->num_charmaps; i++)
 		if (face->charmaps[i]->platform_id == 1 && face->charmaps[i]->encoding_id == 0)
 			if (FT_Get_CMap_Format(face->charmaps[i]) != -1)
+			{
+				fz_ft_unlock(ctx);
 				return face->charmaps[i];
+			}
 
 	if (face->num_charmaps > 0)
 		if (FT_Get_CMap_Format(face->charmaps[0]) != -1)
+		{
+			fz_ft_unlock(ctx);
 			return face->charmaps[0];
+		}
+
+	fz_ft_unlock(ctx);
 	return NULL;
 }
 
@@ -767,7 +789,7 @@ pdf_load_simple_font(fz_context *ctx, pdf_document *doc, pdf_obj *dict)
 	{
 		fontdesc = pdf_new_font_desc(ctx);
 
-		basefont = pdf_to_name(ctx, pdf_dict_get(ctx, dict, PDF_NAME(BaseFont)));
+		basefont = pdf_dict_get_name(ctx, dict, PDF_NAME(BaseFont));
 
 		descriptor = pdf_dict_get(ctx, dict, PDF_NAME(FontDescriptor));
 		if (descriptor)
@@ -805,7 +827,7 @@ pdf_load_simple_font(fz_context *ctx, pdf_document *doc, pdf_obj *dict)
 		}
 
 		face = fontdesc->font->ft_face;
-		kind = ft_kind(face);
+		kind = ft_kind(ctx, face);
 
 		/* Encoding */
 
@@ -817,13 +839,15 @@ pdf_load_simple_font(fz_context *ctx, pdf_document *doc, pdf_obj *dict)
 		if (kind == TYPE1)
 			cmap = select_type1_cmap(face);
 		else if (kind == TRUETYPE)
-			cmap = select_truetype_cmap(face, symbolic);
+			cmap = select_truetype_cmap(ctx, face, symbolic);
 		else
 			cmap = select_unknown_cmap(face);
 
 		if (cmap)
 		{
+			fz_ft_lock(ctx);
 			fterr = FT_Set_Charmap(face, cmap);
+			fz_ft_unlock(ctx);
 			if (fterr)
 				fz_warn(ctx, "freetype could not set cmap: %s", ft_error_string(fterr));
 		}
@@ -874,13 +898,13 @@ pdf_load_simple_font(fz_context *ctx, pdf_document *doc, pdf_obj *dict)
 		else if (!fontdesc->is_embedded && !symbolic)
 			pdf_load_encoding(estrings, "StandardEncoding");
 
+		fz_ft_lock(ctx);
+		has_lock = 1;
+
 		/* start with the builtin encoding */
 		for (i = 0; i < 256; i++)
 			etable[i] = ft_char_index(face, i);
 
-		fz_lock(ctx, FZ_LOCK_FREETYPE);
-		has_lock = 1;
-
 		/* built-in and substitute fonts may be a different type than what the document expects */
 		subtype = pdf_dict_get(ctx, dict, PDF_NAME(Subtype));
 		if (pdf_name_eq(ctx, subtype, PDF_NAME(Type1)))
@@ -987,7 +1011,7 @@ pdf_load_simple_font(fz_context *ctx, pdf_document *doc, pdf_obj *dict)
 					estrings[i] = (char*) fz_glyph_name_from_adobe_standard[i];
 		}
 
-		fz_unlock(ctx, FZ_LOCK_FREETYPE);
+		fz_ft_unlock(ctx);
 		has_lock = 0;
 
 		fontdesc->encoding = pdf_new_identity_cmap(ctx, 0, 1);
@@ -1002,6 +1026,8 @@ pdf_load_simple_font(fz_context *ctx, pdf_document *doc, pdf_obj *dict)
 		fz_catch(ctx)
 		{
 			fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
+			fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
+			fz_report_error(ctx);
 			fz_warn(ctx, "cannot load ToUnicode CMap");
 		}
 
@@ -1030,8 +1056,12 @@ pdf_load_simple_font(fz_context *ctx, pdf_document *doc, pdf_obj *dict)
 		}
 		else
 		{
+			fz_ft_lock(ctx);
+			has_lock = 1;
 			for (i = 0; i < 256; i++)
 				pdf_add_hmtx(ctx, fontdesc, i, i, ft_width(ctx, fontdesc, i));
+			fz_ft_unlock(ctx);
+			has_lock = 0;
 		}
 
 		pdf_end_hmtx(ctx, fontdesc);
@@ -1039,7 +1069,7 @@ pdf_load_simple_font(fz_context *ctx, pdf_document *doc, pdf_obj *dict)
 	fz_catch(ctx)
 	{
 		if (has_lock)
-			fz_unlock(ctx, FZ_LOCK_FREETYPE);
+			fz_ft_unlock(ctx);
 		if (fontdesc && etable != fontdesc->cid_to_gid)
 			fz_free(ctx, etable);
 		pdf_drop_font(ctx, fontdesc);
@@ -1140,7 +1170,7 @@ load_cid_font(fz_context *ctx, pdf_document *doc, pdf_obj *dict, pdf_obj *encodi
 	{
 		/* Get font name and CID collection */
 
-		basefont = pdf_to_name(ctx, pdf_dict_get(ctx, dict, PDF_NAME(BaseFont)));
+		basefont = pdf_dict_get_name(ctx, dict, PDF_NAME(BaseFont));
 
 		{
 			pdf_obj *cidinfo;
@@ -1218,7 +1248,9 @@ load_cid_font(fz_context *ctx, pdf_document *doc, pdf_obj *dict, pdf_obj *encodi
 		/* unicode cmap to get a glyph id */
 		else if (fontdesc->font->flags.ft_substitute)
 		{
+			fz_ft_lock(ctx);
 			fterr = FT_Select_Charmap(face, ft_encoding_unicode);
+			fz_ft_unlock(ctx);
 			if (fterr)
 				fz_throw(ctx, FZ_ERROR_SYNTAX, "no unicode cmap when emulating CID font: %s", ft_error_string(fterr));
 
@@ -1242,17 +1274,33 @@ load_cid_font(fz_context *ctx, pdf_document *doc, pdf_obj *dict, pdf_obj *encodi
 		 * the ToUnicode table if it exists to map via the substitute font's cmap. */
 		if (strstr(fontdesc->encoding->cmap_name, "Identity-") && fontdesc->font->flags.ft_substitute)
 		{
-			fz_warn(ctx, "non-embedded font using identity encoding: %s", basefont);
-			if (fontdesc->to_unicode && !fontdesc->to_ttf_cmap)
-				fontdesc->to_ttf_cmap = pdf_keep_cmap(ctx, fontdesc->to_unicode);
+			if (!fontdesc->to_ttf_cmap)
+			{
+				if (fontdesc->to_unicode)
+				{
+					// Use ToUnicode from PDF file if possible.
+					fontdesc->to_ttf_cmap = pdf_keep_cmap(ctx, fontdesc->to_unicode);
+				}
+				else
+				{
+					// Attempt a generic ToUnicode (default MacRoman ordering for TrueType)
+					fontdesc->to_ttf_cmap = pdf_load_builtin_cmap(ctx, "TrueType-UCS2");
+				}
+			}
+
+			if (fontdesc->to_ttf_cmap)
+			{
+				fz_warn(ctx, "non-embedded font using identity encoding: %s (mapping via %s)", basefont, fontdesc->to_ttf_cmap->cmap_name);
+				if (!fontdesc->to_unicode)
+					fontdesc->to_unicode = pdf_keep_cmap(ctx, fontdesc->to_ttf_cmap);
+			}
+			else
+				fz_warn(ctx, "non-embedded font using identity encoding: %s", basefont);
 		}
 
 		/* Horizontal */
 
-		dw = 1000;
-		obj = pdf_dict_get(ctx, dict, PDF_NAME(DW));
-		if (obj)
-			dw = pdf_to_int(ctx, obj);
+		dw = pdf_dict_get_int_default(ctx, dict, PDF_NAME(DW), 1000);
 		pdf_set_default_hmtx(ctx, fontdesc, dw);
 
 		widths = pdf_dict_get(ctx, dict, PDF_NAME(W));
@@ -1394,8 +1442,7 @@ pdf_load_font_descriptor(fz_context *ctx, pdf_document *doc, pdf_font_desc *font
 
 	fontdesc->flags = pdf_dict_get_int(ctx, dict, PDF_NAME(Flags));
 	fontdesc->italic_angle = pdf_dict_get_real(ctx, dict, PDF_NAME(ItalicAngle));
-	fontdesc->ascent = pdf_dict_get_real(ctx, dict, PDF_NAME(Ascent));
-	fontdesc->descent = pdf_dict_get_real(ctx, dict, PDF_NAME(Descent));
+	/* fontdesc->ascent and descent have already been set to sane defaults */
 	fontdesc->cap_height = pdf_dict_get_real(ctx, dict, PDF_NAME(CapHeight));
 	fontdesc->x_height = pdf_dict_get_real(ctx, dict, PDF_NAME(XHeight));
 	fontdesc->missing_width = pdf_dict_get_real(ctx, dict, PDF_NAME(MissingWidth));
@@ -1414,6 +1461,8 @@ pdf_load_font_descriptor(fz_context *ctx, pdf_document *doc, pdf_font_desc *font
 		fz_catch(ctx)
 		{
 			fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
+			fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
+			fz_report_error(ctx);
 			fz_warn(ctx, "ignored error when loading embedded font; attempting to load system font");
 			if (!iscidfont && fontname != pdf_clean_font_name(fontname))
 				pdf_load_builtin_font(ctx, fontdesc, fontname, 1);
@@ -1431,18 +1480,26 @@ pdf_load_font_descriptor(fz_context *ctx, pdf_document *doc, pdf_font_desc *font
 
 	/* Check for DynaLab fonts that must use hinting */
 	face = fontdesc->font->ft_face;
-	if (ft_kind(face) == TRUETYPE)
+	if (ft_kind(ctx, face) == TRUETYPE)
 	{
 		/* FreeType's own 'tricky' font detection needs a bit of help */
 		if (is_dynalab(fontdesc->font->name))
 			face->face_flags |= FT_FACE_FLAG_TRICKY;
 
-		if (fontdesc->ascent == 0.0f)
-			fontdesc->ascent = 1000.0f * face->ascender / face->units_per_EM;
+		fontdesc->ascent = 1000.0f * face->ascender / face->units_per_EM;
 
-		if (fontdesc->descent == 0.0f)
-			fontdesc->descent = 1000.0f * face->descender / face->units_per_EM;
+		fontdesc->descent = 1000.0f * face->descender / face->units_per_EM;
 	}
+
+	/* Prefer FontDescriptor Ascent/Descent values to embedded font's */
+	fontdesc->ascent = pdf_dict_get_real_default(ctx, dict, PDF_NAME(Ascent), fontdesc->ascent);
+	fontdesc->descent = pdf_dict_get_real_default(ctx, dict, PDF_NAME(Descent), fontdesc->descent);
+	/* Allow for naughty producers that give us a positive descent. */
+	if (fontdesc->descent > 0)
+		fontdesc->descent = -fontdesc->descent;
+
+	fontdesc->font->ascender = fontdesc->ascent / 1000.0f;
+	fontdesc->font->descender = fontdesc->descent / 1000.0f;
 }
 
 static void
@@ -1498,6 +1555,11 @@ pdf_load_font(fz_context *ctx, pdf_document *doc, pdf_obj *rdb, pdf_obj *dict)
 
 	if ((fontdesc = pdf_find_item(ctx, pdf_drop_font_imp, dict)) != NULL)
 	{
+		if (fontdesc->t3loading)
+		{
+			pdf_drop_font(ctx, fontdesc);
+			fz_throw(ctx, FZ_ERROR_SYNTAX, "recursive type3 font");
+		}
 		return fontdesc;
 	}
 
@@ -1540,21 +1602,22 @@ pdf_load_font(fz_context *ctx, pdf_document *doc, pdf_obj *rdb, pdf_obj *dict)
 		/* Create glyph width table for stretching substitute fonts and text extraction. */
 		pdf_make_width_table(ctx, fontdesc);
 
+		pdf_store_item(ctx, dict, fontdesc, fontdesc->size);
+
 		/* Load CharProcs */
 		if (type3)
 		{
-			if (doc->type3_lock)
-				fz_throw(ctx, FZ_ERROR_GENERIC, "recursive type3 font");
-			doc->type3_lock = 1;
+			fontdesc->t3loading = 1;
 			fz_try(ctx)
 				pdf_load_type3_glyphs(ctx, doc, fontdesc);
 			fz_always(ctx)
-				doc->type3_lock = 0;
+				fontdesc->t3loading = 0;
 			fz_catch(ctx)
+			{
+				pdf_remove_item(ctx, fontdesc->storable.drop, dict);
 				fz_rethrow(ctx);
+			}
 		}
-
-		pdf_store_item(ctx, dict, fontdesc, fontdesc->size);
 	}
 	fz_catch(ctx)
 	{
