@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2022 Artifex Software, Inc.
+// Copyright (C) 2004-2024 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -508,6 +508,9 @@ pdf_obj *pdf_button_field_on_state(fz_context *ctx, pdf_obj *field)
 static void
 begin_annot_op(fz_context *ctx, pdf_annot *annot, const char *op)
 {
+	if (!annot->page)
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "annotation not bound to any page");
+
 	pdf_begin_operation(ctx, annot->page->doc, op);
 }
 
@@ -1009,7 +1012,12 @@ pdf_annot *
 pdf_create_signature_widget(fz_context *ctx, pdf_page *page, char *name)
 {
 	fz_rect rect = { 12, 12, 12+100, 12+50 };
-	pdf_annot *annot = pdf_create_annot_raw(ctx, page, PDF_ANNOT_WIDGET);
+	pdf_annot *annot;
+
+	pdf_begin_operation(ctx, page->doc, "Create signature");
+
+	annot = pdf_create_annot_raw(ctx, page, PDF_ANNOT_WIDGET);
+
 	fz_try(ctx)
 	{
 		pdf_obj *obj = annot->obj;
@@ -1035,9 +1043,11 @@ pdf_create_signature_widget(fz_context *ctx, pdf_page *page, char *name)
 		pdf_array_push(ctx, fields, obj);
 		lock = pdf_dict_put_dict(ctx, obj, PDF_NAME(Lock), 1);
 		pdf_dict_put(ctx, lock, PDF_NAME(Action), PDF_NAME(All));
+		pdf_end_operation(ctx, page->doc);
 	}
 	fz_catch(ctx)
 	{
+		pdf_abandon_operation(ctx, page->doc);
 		pdf_delete_annot(ctx, page, annot);
 	}
 	return (pdf_annot *)annot;
@@ -1116,14 +1126,15 @@ merge_changes(fz_context *ctx, const char *value, int start, int end, const char
 
 int pdf_set_text_field_value(fz_context *ctx, pdf_annot *widget, const char *update)
 {
-	pdf_document *doc = widget->page->doc;
+	pdf_document *doc;
 	pdf_keystroke_event evt = { 0 };
 	char *new_change = NULL;
 	char *new_value = NULL;
 	char *merged_value = NULL;
 	int rc = 1;
 
-	pdf_begin_operation(ctx, doc, "Edit text field");
+	begin_annot_op(ctx, widget, "Edit text field");
+	doc = widget->page->doc;
 
 	fz_var(new_value);
 	fz_var(new_change);
@@ -1159,7 +1170,7 @@ int pdf_set_text_field_value(fz_context *ctx, pdf_annot *widget, const char *upd
 		{
 			rc = pdf_set_annot_field_value(ctx, doc, widget, update, 1);
 		}
-		pdf_end_operation(ctx, doc);
+		end_annot_op(ctx, widget);
 	}
 	fz_always(ctx)
 	{
@@ -1171,7 +1182,7 @@ int pdf_set_text_field_value(fz_context *ctx, pdf_annot *widget, const char *upd
 	}
 	fz_catch(ctx)
 	{
-		pdf_abandon_operation(ctx, doc);
+		abandon_annot_op(ctx, widget);
 		fz_warn(ctx, "could not set widget text");
 		rc = 0;
 	}
@@ -1507,12 +1518,24 @@ fz_stream *pdf_signature_hash_bytes(fz_context *ctx, pdf_document *doc, pdf_obj
 	return bytes;
 }
 
+int pdf_incremental_change_since_signing_widget(fz_context *ctx, pdf_annot *widget)
+{
+	if (!widget->page)
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "annotation not bound to any page");
+	return pdf_signature_incremental_change_since_signing(ctx, widget->page->doc, widget->obj);
+}
+
 int pdf_signature_incremental_change_since_signing(fz_context *ctx, pdf_document *doc, pdf_obj *signature)
 {
 	fz_range *byte_range = NULL;
 	int byte_range_len;
 	int changed = 0;
 
+	if (pdf_dict_get_inheritable(ctx, signature, PDF_NAME(FT)) != PDF_NAME(Sig))
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "annotation is not a signature widget");
+	if (!pdf_signature_is_signed(ctx, doc, signature))
+		return 0;
+
 	fz_var(byte_range);
 	fz_try(ctx)
 	{
@@ -1567,6 +1590,10 @@ int pdf_widget_is_signed(fz_context *ctx, pdf_annot *widget)
 {
 	if (widget == NULL)
 		return 0;
+
+	if (!widget->page)
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "annotation not bound to any page");
+
 	return pdf_signature_is_signed(ctx, widget->page->doc, widget->obj);
 }
 
@@ -2367,7 +2394,7 @@ static void pdf_bake_page(fz_context *ctx, pdf_document *doc, pdf_obj *page, int
 	pdf_obj *subtype;
 	pdf_obj *prologue = NULL;
 	fz_buffer *buf = NULL;
-	int underflow, overflow;
+	int prepend, append;
 	int i;
 
 	fz_var(buf);
@@ -2390,22 +2417,23 @@ static void pdf_bake_page(fz_context *ctx, pdf_document *doc, pdf_obj *page, int
 	{
 		// Ensure that the graphics state is balanced.
 		contents = pdf_dict_get(ctx, page, PDF_NAME(Contents));
-		pdf_count_q_balance(ctx, doc, res, contents, &underflow, &overflow);
+		pdf_count_q_balance(ctx, doc, res, contents, &prepend, &append);
 
-		// Prepend enough 'q' to ensure we can get back to initial state.
-		buf = fz_new_buffer(ctx, 1024);
-		fz_append_string(ctx, buf, "q\n");
-		while (underflow-- > 0)
-			fz_append_string(ctx, buf, "q\n");
+		if (prepend)
+		{
+			// Prepend enough 'q' to ensure we can get back to initial state.
+			buf = fz_new_buffer(ctx, 1024);
+			while (prepend-- > 0)
+				fz_append_string(ctx, buf, "q\n");
 
-		prologue = pdf_add_stream(ctx, doc, buf, NULL, 0);
-		fz_drop_buffer(ctx, buf);
-		buf = NULL;
+			prologue = pdf_add_stream(ctx, doc, buf, NULL, 0);
+			fz_drop_buffer(ctx, buf);
+			buf = NULL;
+		}
 
 		// Append enough 'Q' to get back to initial state.
 		buf = fz_new_buffer(ctx, 1024);
-		fz_append_string(ctx, buf, "Q\n");
-		while (overflow-- > 0)
+		while (append-- > 0)
 			fz_append_string(ctx, buf, "Q\n");
 
 		for (i = 0; i < pdf_array_len(ctx, annots); )
@@ -2445,14 +2473,16 @@ static void pdf_bake_page(fz_context *ctx, pdf_document *doc, pdf_obj *page, int
 		if (!pdf_is_array(ctx, contents))
 		{
 			new_contents = pdf_new_array(ctx, doc, 10);
-			pdf_array_push(ctx, new_contents, prologue);
-			pdf_array_push(ctx, new_contents, contents);
+			if (prologue)
+				pdf_array_push(ctx, new_contents, prologue);
+			if (contents)
+				pdf_array_push(ctx, new_contents, contents);
 			pdf_dict_put(ctx, page, PDF_NAME(Contents), new_contents);
 			pdf_drop_obj(ctx, new_contents);
 			contents = new_contents;
 			new_contents = NULL;
 		}
-		else
+		else if (prologue)
 		{
 			pdf_array_insert(ctx, contents, prologue, 0);
 		}
