@@ -1,6 +1,27 @@
+// Copyright (C) 2004-2024 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
+
 #include "mupdf/fitz.h"
 #include "mupdf/pdf.h"
-#include "pdf-imp.h"
 
 #include <string.h>
 
@@ -12,7 +33,7 @@ struct entry
 	int gen;
 	int64_t ofs;
 	int64_t stm_ofs;
-	int stm_len;
+	int64_t stm_len;
 };
 
 static void add_root(fz_context *ctx, pdf_obj *obj, pdf_obj ***roots, int *num_roots, int *max_roots)
@@ -22,18 +43,24 @@ static void add_root(fz_context *ctx, pdf_obj *obj, pdf_obj ***roots, int *num_r
 		int new_max_roots = *max_roots * 2;
 		if (new_max_roots == 0)
 			new_max_roots = 4;
-		*roots = fz_resize_array(ctx, *roots, new_max_roots, sizeof(**roots));
+		*roots = fz_realloc_array(ctx, *roots, new_max_roots, pdf_obj*);
 		*max_roots = new_max_roots;
 	}
 	(*roots)[(*num_roots)++] = pdf_keep_obj(ctx, obj);
 }
 
 int
-pdf_repair_obj(fz_context *ctx, pdf_document *doc, pdf_lexbuf *buf, int64_t *stmofsp, int *stmlenp, pdf_obj **encrypt, pdf_obj **id, pdf_obj **page, int64_t *tmpofs, pdf_obj **root)
+pdf_repair_obj(fz_context *ctx, pdf_document *doc, pdf_lexbuf *buf, int64_t *stmofsp, int64_t *stmlenp, pdf_obj **encrypt, pdf_obj **id, pdf_obj **page, int64_t *tmpofs, pdf_obj **root)
 {
 	fz_stream *file = doc->file;
 	pdf_token tok;
-	int stm_len;
+	int64_t stm_len;
+	int64_t local_ofs;
+
+	if (tmpofs == NULL)
+		tmpofs = &local_ofs;
+	if (stmofsp == NULL)
+		stmofsp = &local_ofs;
 
 	*stmofsp = 0;
 	if (stmlenp)
@@ -41,12 +68,20 @@ pdf_repair_obj(fz_context *ctx, pdf_document *doc, pdf_lexbuf *buf, int64_t *stm
 
 	stm_len = 0;
 
+	*tmpofs = fz_tell(ctx, file);
+	if (*tmpofs < 0)
+		fz_throw(ctx, FZ_ERROR_SYSTEM, "cannot tell in file");
+
 	/* On entry to this function, we know that we've just seen
 	 * '<int> <int> obj'. We expect the next thing we see to be a
 	 * pdf object. Regardless of the type of thing we meet next
 	 * we only need to fully parse it if it is a dictionary. */
 	tok = pdf_lex(ctx, file, buf);
 
+	/* Don't let a truncated object at EOF overwrite a good one */
+	if (tok == PDF_TOK_EOF)
+		fz_throw(ctx, FZ_ERROR_SYNTAX, "truncated object");
+
 	if (tok == PDF_TOK_OPEN_DICT)
 	{
 		pdf_obj *obj, *dict = NULL;
@@ -58,11 +93,13 @@ pdf_repair_obj(fz_context *ctx, pdf_document *doc, pdf_lexbuf *buf, int64_t *stm
 		fz_catch(ctx)
 		{
 			fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
+			fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
 			/* Don't let a broken object at EOF overwrite a good one */
 			if (file->eof)
 				fz_rethrow(ctx);
 			/* Silently swallow the error */
-			dict = pdf_new_dict(ctx, NULL, 2);
+			fz_report_error(ctx);
+			dict = pdf_new_dict(ctx, doc, 2);
 		}
 
 		/* We must be careful not to try to resolve any indirections
@@ -73,12 +110,12 @@ pdf_repair_obj(fz_context *ctx, pdf_document *doc, pdf_lexbuf *buf, int64_t *stm
 
 		if (encrypt || id || root)
 		{
-			obj = pdf_dict_get(ctx, dict, PDF_NAME_Type);
-			if (!pdf_is_indirect(ctx, obj) && pdf_name_eq(ctx, obj, PDF_NAME_XRef))
+			obj = pdf_dict_get(ctx, dict, PDF_NAME(Type));
+			if (!pdf_is_indirect(ctx, obj) && pdf_name_eq(ctx, obj, PDF_NAME(XRef)))
 			{
 				if (encrypt)
 				{
-					obj = pdf_dict_get(ctx, dict, PDF_NAME_Encrypt);
+					obj = pdf_dict_get(ctx, dict, PDF_NAME(Encrypt));
 					if (obj)
 					{
 						pdf_drop_obj(ctx, *encrypt);
@@ -88,7 +125,7 @@ pdf_repair_obj(fz_context *ctx, pdf_document *doc, pdf_lexbuf *buf, int64_t *stm
 
 				if (id)
 				{
-					obj = pdf_dict_get(ctx, dict, PDF_NAME_ID);
+					obj = pdf_dict_get(ctx, dict, PDF_NAME(ID));
 					if (obj)
 					{
 						pdf_drop_obj(ctx, *id);
@@ -97,18 +134,18 @@ pdf_repair_obj(fz_context *ctx, pdf_document *doc, pdf_lexbuf *buf, int64_t *stm
 				}
 
 				if (root)
-					*root = pdf_keep_obj(ctx, pdf_dict_get(ctx, dict, PDF_NAME_Root));
+					*root = pdf_keep_obj(ctx, pdf_dict_get(ctx, dict, PDF_NAME(Root)));
 			}
 		}
 
-		obj = pdf_dict_get(ctx, dict, PDF_NAME_Length);
+		obj = pdf_dict_get(ctx, dict, PDF_NAME(Length));
 		if (!pdf_is_indirect(ctx, obj) && pdf_is_int(ctx, obj))
-			stm_len = pdf_to_int(ctx, obj);
+			stm_len = pdf_to_int64(ctx, obj);
 
 		if (doc->file_reading_linearly && page)
 		{
-			obj = pdf_dict_get(ctx, dict, PDF_NAME_Type);
-			if (!pdf_is_indirect(ctx, obj) && pdf_name_eq(ctx, obj, PDF_NAME_Page))
+			obj = pdf_dict_get(ctx, dict, PDF_NAME(Type));
+			if (!pdf_is_indirect(ctx, obj) && pdf_name_eq(ctx, obj, PDF_NAME(Page)))
 			{
 				pdf_drop_obj(ctx, *page);
 				*page = pdf_keep_obj(ctx, dict);
@@ -126,7 +163,7 @@ pdf_repair_obj(fz_context *ctx, pdf_document *doc, pdf_lexbuf *buf, int64_t *stm
 	{
 		*tmpofs = fz_tell(ctx, file);
 		if (*tmpofs < 0)
-			fz_throw(ctx, FZ_ERROR_GENERIC, "cannot tell in file");
+			fz_throw(ctx, FZ_ERROR_SYSTEM, "cannot tell in file");
 		tok = pdf_lex(ctx, file, buf);
 	}
 
@@ -141,7 +178,7 @@ pdf_repair_obj(fz_context *ctx, pdf_document *doc, pdf_lexbuf *buf, int64_t *stm
 
 		*stmofsp = fz_tell(ctx, file);
 		if (*stmofsp < 0)
-			fz_throw(ctx, FZ_ERROR_GENERIC, "cannot seek in file");
+			fz_throw(ctx, FZ_ERROR_SYSTEM, "cannot tell in file");
 
 		if (stm_len > 0)
 		{
@@ -153,6 +190,8 @@ pdf_repair_obj(fz_context *ctx, pdf_document *doc, pdf_lexbuf *buf, int64_t *stm
 			fz_catch(ctx)
 			{
 				fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
+				fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
+				fz_report_error(ctx);
 				fz_warn(ctx, "cannot find endstream token, falling back to scanning");
 			}
 			if (tok == PDF_TOK_ENDSTREAM)
@@ -177,7 +216,7 @@ pdf_repair_obj(fz_context *ctx, pdf_document *doc, pdf_lexbuf *buf, int64_t *stm
 atobjend:
 		*tmpofs = fz_tell(ctx, file);
 		if (*tmpofs < 0)
-			fz_throw(ctx, FZ_ERROR_GENERIC, "cannot tell in file");
+			fz_throw(ctx, FZ_ERROR_SYSTEM, "cannot tell in file");
 		tok = pdf_lex(ctx, file, buf);
 		if (tok != PDF_TOK_ENDOBJ)
 			fz_warn(ctx, "object missing 'endobj' token");
@@ -186,7 +225,7 @@ atobjend:
 			/* Read another token as we always return the next one */
 			*tmpofs = fz_tell(ctx, file);
 			if (*tmpofs < 0)
-				fz_throw(ctx, FZ_ERROR_GENERIC, "cannot tell in file");
+				fz_throw(ctx, FZ_ERROR_SYSTEM, "cannot tell in file");
 			tok = pdf_lex(ctx, file, buf);
 		}
 	}
@@ -210,7 +249,7 @@ pdf_repair_obj_stm(fz_context *ctx, pdf_document *doc, int stm_num)
 	{
 		obj = pdf_load_object(ctx, doc, stm_num);
 
-		count = pdf_to_int(ctx, pdf_dict_get(ctx, obj, PDF_NAME_N));
+		count = pdf_dict_get_int(ctx, obj, PDF_NAME(N));
 
 		pdf_drop_obj(ctx, obj);
 
@@ -222,7 +261,7 @@ pdf_repair_obj_stm(fz_context *ctx, pdf_document *doc, int stm_num)
 
 			tok = pdf_lex(ctx, stm, &buf);
 			if (tok != PDF_TOK_INT)
-				fz_throw(ctx, FZ_ERROR_GENERIC, "corrupt object stream (%d 0 R)", stm_num);
+				fz_throw(ctx, FZ_ERROR_FORMAT, "corrupt object stream (%d 0 R)", stm_num);
 
 			n = buf.i;
 			if (n < 0)
@@ -230,7 +269,7 @@ pdf_repair_obj_stm(fz_context *ctx, pdf_document *doc, int stm_num)
 				fz_warn(ctx, "ignoring object with invalid object number (%d %d R)", n, i);
 				continue;
 			}
-			else if (n >= pdf_xref_len(ctx, doc))
+			else if (n >= PDF_MAX_OBJECT_NUMBER)
 			{
 				fz_warn(ctx, "ignoring object with invalid object number (%d %d R)", n, i);
 				continue;
@@ -247,7 +286,7 @@ pdf_repair_obj_stm(fz_context *ctx, pdf_document *doc, int stm_num)
 
 			tok = pdf_lex(ctx, stm, &buf);
 			if (tok != PDF_TOK_INT)
-				fz_throw(ctx, FZ_ERROR_GENERIC, "corrupt object stream (%d 0 R)", stm_num);
+				fz_throw(ctx, FZ_ERROR_FORMAT, "corrupt object stream (%d 0 R)", stm_num);
 		}
 	}
 	fz_always(ctx)
@@ -270,7 +309,7 @@ orphan_object(fz_context *ctx, pdf_document *doc, pdf_obj *obj)
 
 		fz_try(ctx)
 		{
-			doc->orphans = fz_resize_array(ctx, doc->orphans, new_max, sizeof(*doc->orphans));
+			doc->orphans = fz_realloc_array(ctx, doc->orphans, new_max, pdf_obj*);
 			doc->orphans_max = new_max;
 		}
 		fz_catch(ctx)
@@ -306,7 +345,7 @@ pdf_repair_xref(fz_context *ctx, pdf_document *doc)
 	int num = 0;
 	int gen = 0;
 	int64_t tmpofs, stm_ofs, numofs = 0, genofs = 0;
-	int stm_len;
+	int64_t stm_len;
 	pdf_token tok;
 	int next;
 	int i;
@@ -325,15 +364,17 @@ pdf_repair_xref(fz_context *ctx, pdf_document *doc)
 	fz_var(list);
 	fz_var(obj);
 
-	fz_warn(ctx, "repairing PDF document");
+	if (!doc->is_fdf)
+		fz_warn(ctx, "repairing PDF document");
 
 	if (doc->repair_attempted)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "Repair failed already - not trying again");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "Repair failed already - not trying again");
+
 	doc->repair_attempted = 1;
+	doc->repair_in_progress = 1;
 
-	doc->dirty = 1;
-	doc->freeze_updates = 1; /* Can't support incremental update after repair */
-
+	pdf_drop_page_tree_internal(ctx, doc);
+	doc->page_tree_broken = 0;
 	pdf_forget_xref(ctx, doc);
 
 	fz_seek(ctx, doc->file, 0, 0);
@@ -343,17 +384,17 @@ pdf_repair_xref(fz_context *ctx, pdf_document *doc)
 		pdf_xref_entry *entry;
 		listlen = 0;
 		listcap = 1024;
-		list = fz_malloc_array(ctx, listcap, sizeof(struct entry));
+		list = fz_malloc_array(ctx, listcap, struct entry);
 
 		/* look for '%PDF' version marker within first kilobyte of file */
-		n = fz_read(ctx, doc->file, (unsigned char *)buf->scratch, fz_mini(buf->size, 1024));
+		n = fz_read(ctx, doc->file, (unsigned char *)buf->scratch, fz_minz(buf->size, 1024));
 
 		fz_seek(ctx, doc->file, 0, 0);
 		if (n >= 4)
 		{
 			for (j = 0; j < n - 4; j++)
 			{
-				if (memcmp(&buf->scratch[j], "%PDF", 4) == 0)
+				if (memcmp(&buf->scratch[j], "%PDF", 4) == 0 || memcmp(&buf->scratch[j], "%FDF", 4) == 0)
 				{
 					fz_seek(ctx, doc->file, (int64_t)(j + 8), 0); /* skip "%PDF-X.Y" */
 					break;
@@ -366,24 +407,30 @@ pdf_repair_xref(fz_context *ctx, pdf_document *doc)
 		c = fz_read_byte(ctx, doc->file);
 		while (c >= 0 && (c == ' ' || c == '%'))
 			c = fz_read_byte(ctx, doc->file);
-		fz_unread_byte(ctx, doc->file);
+		if (c != EOF)
+			fz_unread_byte(ctx, doc->file);
 
 		while (1)
 		{
 			tmpofs = fz_tell(ctx, doc->file);
 			if (tmpofs < 0)
-				fz_throw(ctx, FZ_ERROR_GENERIC, "cannot tell in file");
+				fz_throw(ctx, FZ_ERROR_SYSTEM, "cannot tell in file");
 
 			fz_try(ctx)
 				tok = pdf_lex_no_string(ctx, doc->file, buf);
 			fz_catch(ctx)
 			{
 				fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
+				fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
+				fz_report_error(ctx);
 				fz_warn(ctx, "skipping ahead to next token");
 				do
 					c = fz_read_byte(ctx, doc->file);
 				while (c != EOF && !is_white(c));
-				continue;
+				if (c == EOF)
+					tok = PDF_TOK_EOF;
+				else
+					continue;
 			}
 
 			/* If we have the next token already, then we'll jump
@@ -424,11 +471,13 @@ pdf_repair_xref(fz_context *ctx, pdf_document *doc)
 				fz_catch(ctx)
 				{
 					fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
+					fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
 					/* If we haven't seen a root yet, there is nothing
 					 * we can do, but give up. Otherwise, we'll make
 					 * do. */
 					if (!roots)
 						fz_rethrow(ctx);
+					fz_report_error(ctx);
 					fz_warn(ctx, "cannot parse object (%d %d R) - ignoring rest of file", num, gen);
 					break;
 				}
@@ -444,7 +493,7 @@ pdf_repair_xref(fz_context *ctx, pdf_document *doc)
 				if (listlen + 1 == listcap)
 				{
 					listcap = (listcap * 3) / 2;
-					list = fz_resize_array(ctx, list, listcap, sizeof(struct entry));
+					list = fz_realloc_array(ctx, list, listcap, struct entry);
 				}
 
 				list[listlen].num = num;
@@ -465,6 +514,8 @@ pdf_repair_xref(fz_context *ctx, pdf_document *doc)
 			 * by a corrupt file. */
 			else if (tok == PDF_TOK_OPEN_DICT)
 			{
+				pdf_obj *dictobj;
+
 				fz_try(ctx)
 				{
 					dict = pdf_parse_dict(ctx, doc, doc->file, buf);
@@ -472,46 +523,46 @@ pdf_repair_xref(fz_context *ctx, pdf_document *doc)
 				fz_catch(ctx)
 				{
 					fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
+					fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
 					/* If this was the real trailer dict
 					 * it was broken, in which case we are
 					 * in trouble. Keep going though in
 					 * case this was just a bogus dict. */
+					fz_report_error(ctx);
 					continue;
 				}
 
 				fz_try(ctx)
 				{
-					obj = pdf_dict_get(ctx, dict, PDF_NAME_Encrypt);
-					if (obj)
+					dictobj = pdf_dict_get(ctx, dict, PDF_NAME(Encrypt));
+					if (dictobj)
 					{
 						pdf_drop_obj(ctx, encrypt);
-						encrypt = pdf_keep_obj(ctx, obj);
+						encrypt = pdf_keep_obj(ctx, dictobj);
 					}
 
-					obj = pdf_dict_get(ctx, dict, PDF_NAME_ID);
-					if (obj && (!id || !encrypt || pdf_dict_get(ctx, dict, PDF_NAME_Encrypt)))
+					dictobj = pdf_dict_get(ctx, dict, PDF_NAME(ID));
+					if (dictobj && (!id || !encrypt || pdf_dict_get(ctx, dict, PDF_NAME(Encrypt))))
 					{
 						pdf_drop_obj(ctx, id);
-						id = pdf_keep_obj(ctx, obj);
+						id = pdf_keep_obj(ctx, dictobj);
 					}
 
-					obj = pdf_dict_get(ctx, dict, PDF_NAME_Root);
-					if (obj)
-						add_root(ctx, obj, &roots, &num_roots, &max_roots);
+					dictobj = pdf_dict_get(ctx, dict, PDF_NAME(Root));
+					if (dictobj)
+						add_root(ctx, dictobj, &roots, &num_roots, &max_roots);
 
-					obj = pdf_dict_get(ctx, dict, PDF_NAME_Info);
-					if (obj)
+					dictobj = pdf_dict_get(ctx, dict, PDF_NAME(Info));
+					if (dictobj)
 					{
 						pdf_drop_obj(ctx, info);
-						info = pdf_keep_obj(ctx, obj);
+						info = pdf_keep_obj(ctx, dictobj);
 					}
 				}
 				fz_always(ctx)
 					pdf_drop_obj(ctx, dict);
 				fz_catch(ctx)
 					fz_rethrow(ctx);
-
-				obj = NULL;
 			}
 
 			else if (tok == PDF_TOK_EOF)
@@ -527,7 +578,7 @@ pdf_repair_xref(fz_context *ctx, pdf_document *doc)
 		}
 
 		if (listlen == 0)
-			fz_throw(ctx, FZ_ERROR_GENERIC, "no objects found");
+			fz_throw(ctx, FZ_ERROR_FORMAT, "no objects found");
 
 		/* make xref reasonable */
 
@@ -570,8 +621,8 @@ pdf_repair_xref(fz_context *ctx, pdf_document *doc)
 
 				fz_try(ctx)
 				{
-					length = pdf_new_int(ctx, doc, list[i].stm_len);
-					pdf_dict_get_put_drop(ctx, dict, PDF_NAME_Length, length, &old_obj);
+					length = pdf_new_int(ctx, list[i].stm_len);
+					pdf_dict_get_put_drop(ctx, dict, PDF_NAME(Length), length, &old_obj);
 					if (old_obj)
 						orphan_object(ctx, doc, old_obj);
 				}
@@ -610,10 +661,7 @@ pdf_repair_xref(fz_context *ctx, pdf_document *doc)
 		pdf_drop_obj(ctx, obj);
 		obj = NULL;
 
-		obj = pdf_new_int(ctx, doc, maxnum + 1);
-		pdf_dict_put(ctx, pdf_trailer(ctx, doc), PDF_NAME_Size, obj);
-		pdf_drop_obj(ctx, obj);
-		obj = NULL;
+		pdf_dict_put_int(ctx, pdf_trailer(ctx, doc), PDF_NAME(Size), maxnum + 1);
 
 		if (roots)
 		{
@@ -624,12 +672,12 @@ pdf_repair_xref(fz_context *ctx, pdf_document *doc)
 			}
 			if (i >= 0)
 			{
-				pdf_dict_put(ctx, pdf_trailer(ctx, doc), PDF_NAME_Root, roots[i]);
+				pdf_dict_put(ctx, pdf_trailer(ctx, doc), PDF_NAME(Root), roots[i]);
 			}
 		}
 		if (info)
 		{
-			pdf_dict_put(ctx, pdf_trailer(ctx, doc), PDF_NAME_Info, info);
+			pdf_dict_put(ctx, pdf_trailer(ctx, doc), PDF_NAME(Info), info);
 			pdf_drop_obj(ctx, info);
 			info = NULL;
 		}
@@ -644,7 +692,7 @@ pdf_repair_xref(fz_context *ctx, pdf_document *doc)
 				encrypt = obj;
 				obj = NULL;
 			}
-			pdf_dict_put(ctx, pdf_trailer(ctx, doc), PDF_NAME_Encrypt, encrypt);
+			pdf_dict_put(ctx, pdf_trailer(ctx, doc), PDF_NAME(Encrypt), encrypt);
 			pdf_drop_obj(ctx, encrypt);
 			encrypt = NULL;
 		}
@@ -659,18 +707,18 @@ pdf_repair_xref(fz_context *ctx, pdf_document *doc)
 				id = obj;
 				obj = NULL;
 			}
-			pdf_dict_put(ctx, pdf_trailer(ctx, doc), PDF_NAME_ID, id);
+			pdf_dict_put(ctx, pdf_trailer(ctx, doc), PDF_NAME(ID), id);
 			pdf_drop_obj(ctx, id);
 			id = NULL;
 		}
-
-		fz_free(ctx, list);
 	}
 	fz_always(ctx)
 	{
 		for (i = 0; i < num_roots; i++)
 			pdf_drop_obj(ctx, roots[i]);
 		fz_free(ctx, roots);
+		fz_free(ctx, list);
+		doc->repair_in_progress = 0;
 	}
 	fz_catch(ctx)
 	{
@@ -678,9 +726,13 @@ pdf_repair_xref(fz_context *ctx, pdf_document *doc)
 		pdf_drop_obj(ctx, id);
 		pdf_drop_obj(ctx, obj);
 		pdf_drop_obj(ctx, info);
-		fz_free(ctx, list);
+		if (ctx->throw_on_repair)
+			fz_throw(ctx, FZ_ERROR_REPAIRED, "Error during repair attempt");
 		fz_rethrow(ctx);
 	}
+
+	if (ctx->throw_on_repair)
+		fz_throw(ctx, FZ_ERROR_REPAIRED, "File repaired");
 }
 
 void
@@ -699,14 +751,17 @@ pdf_repair_obj_stms(fz_context *ctx, pdf_document *doc)
 			dict = pdf_load_object(ctx, doc, i);
 			fz_try(ctx)
 			{
-				if (pdf_name_eq(ctx, pdf_dict_get(ctx, dict, PDF_NAME_Type), PDF_NAME_ObjStm))
+				if (pdf_name_eq(ctx, pdf_dict_get(ctx, dict, PDF_NAME(Type)), PDF_NAME(ObjStm)))
 					pdf_repair_obj_stm(ctx, doc, i);
 			}
+			fz_always(ctx)
+				pdf_drop_obj(ctx, dict);
 			fz_catch(ctx)
 			{
+				fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
+				fz_report_error(ctx);
 				fz_warn(ctx, "ignoring broken object stream (%d 0 R)", i);
 			}
-			pdf_drop_obj(ctx, dict);
 		}
 	}
 
@@ -716,6 +771,9 @@ pdf_repair_obj_stms(fz_context *ctx, pdf_document *doc)
 		pdf_xref_entry *entry = pdf_get_populating_xref_entry(ctx, doc, i);
 
 		if (entry->type == 'o' && pdf_get_populating_xref_entry(ctx, doc, entry->ofs)->type != 'n')
-			fz_throw(ctx, FZ_ERROR_GENERIC, "invalid reference to non-object-stream: %d (%d 0 R)", (int)entry->ofs, i);
+		{
+			fz_warn(ctx, "invalid reference to non-object-stream: %d, assuming %d 0 R is a freed object", (int)entry->ofs, i);
+			entry->type = 'f';
+		}
 	}
 }
