@@ -46,7 +46,7 @@ static const char *options = "";
 static fz_context *ctx;
 static fz_document *doc;
 static fz_document_writer *out;
-static fz_box_type page_box = FZ_MEDIA_BOX;
+static fz_box_type page_box = FZ_CROP_BOX;
 static int count;
 
 static int usage(void)
@@ -182,7 +182,8 @@ int muconvert_main(int argc, char **argv)
 		fz_register_document_handlers(ctx);
 	fz_catch(ctx)
 	{
-		fprintf(stderr, "cannot register document handlers: %s\n", fz_caught_message(ctx));
+		fz_report_error(ctx);
+		fprintf(stderr, "cannot register document handlers\n");
 		fz_drop_context(ctx);
 		return EXIT_FAILURE;
 	}
@@ -190,11 +191,7 @@ int muconvert_main(int argc, char **argv)
 	fz_set_aa_level(ctx, alphabits);
 
 	if (layout_css)
-	{
-		fz_buffer *buf = fz_read_file(ctx, layout_css);
-		fz_set_user_css(ctx, fz_string_from_buffer(ctx, buf));
-		fz_drop_buffer(ctx, buf);
-	}
+		fz_load_user_css(ctx, layout_css);
 
 	fz_set_use_document_css(ctx, layout_use_doc_css);
 
@@ -203,11 +200,13 @@ int muconvert_main(int argc, char **argv)
 		out = fz_new_document_writer(ctx, output, format, options);
 	fz_catch(ctx)
 	{
-		fprintf(stderr, "cannot create document: %s\n", fz_caught_message(ctx));
+		fz_report_error(ctx);
+		fprintf(stderr, "cannot create document\n");
 		fz_drop_context(ctx);
 		return EXIT_FAILURE;
 	}
 
+	fz_var(doc);
 	fz_try(ctx)
 	{
 		for (i = fz_optind; i < argc; ++i)
@@ -215,7 +214,7 @@ int muconvert_main(int argc, char **argv)
 			doc = fz_open_document(ctx, argv[i]);
 			if (fz_needs_password(ctx, doc))
 				if (!fz_authenticate_password(ctx, doc, password))
-					fz_throw(ctx, FZ_ERROR_GENERIC, "cannot authenticate password: %s", argv[i]);
+					fz_throw(ctx, FZ_ERROR_ARGUMENT, "cannot authenticate password: %s", argv[i]);
 			fz_layout_document(ctx, doc, layout_w, layout_h, layout_em);
 			count = fz_count_pages(ctx, doc);
 
@@ -236,7 +235,7 @@ int muconvert_main(int argc, char **argv)
 	}
 	fz_catch(ctx)
 	{
-		fz_log_error(ctx, fz_caught_message(ctx));
+		fz_report_error(ctx);
 		retval = EXIT_FAILURE;
 	}
 
