@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2023 Artifex Software, Inc.
+// Copyright (C) 2004-2024 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -17,8 +17,8 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 /*
  * muraster -- Convert a document to a raster file.
@@ -165,7 +165,7 @@
 #include <stdlib.h>
 #include <stdio.h>
 
-#ifdef _MSC_VER
+#ifdef _WIN32
 struct timeval;
 struct timezone;
 int gettimeofday(struct timeval *tv, struct timezone *tz);
@@ -1005,7 +1005,7 @@ initialise_banding(fz_context *ctx, render_details *render, int color)
 	w = render->ibounds.x1 - render->ibounds.x0;
 	h = render->ibounds.y1 - render->ibounds.y0;
 	if (w <= 0 || h <= 0)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "Invalid page dimensions");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "Invalid page dimensions");
 
 
 	min_band_mem = (size_t)bpp * w * min_band_height;
@@ -1608,11 +1608,7 @@ int main(int argc, char **argv)
 #endif /* DISABLE_MUTHREADS */
 
 	if (layout_css)
-	{
-		fz_buffer *buf = fz_read_file(ctx, layout_css);
-		fz_set_user_css(ctx, fz_string_from_buffer(ctx, buf));
-		fz_drop_buffer(ctx, buf);
-	}
+		fz_load_user_css(ctx, layout_css);
 
 	fz_set_use_document_css(ctx, layout_use_doc_css);
 
@@ -1702,7 +1698,7 @@ int main(int argc, char **argv)
 				if (fz_needs_password(ctx, doc))
 				{
 					if (!fz_authenticate_password(ctx, doc, password))
-						fz_throw(ctx, FZ_ERROR_GENERIC, "cannot authenticate password: %s", filename);
+						fz_throw(ctx, FZ_ERROR_ARGUMENT, "cannot authenticate password: %s", filename);
 				}
 
 				fz_layout_document(ctx, doc, layout_w, layout_h, layout_em);
@@ -1722,6 +1718,7 @@ int main(int argc, char **argv)
 
 				fz_drop_document(ctx, doc);
 				doc = NULL;
+				fz_report_error(ctx);
 				fz_warn(ctx, "ignoring error in '%s'", filename);
 			}
 		}
@@ -1730,7 +1727,7 @@ int main(int argc, char **argv)
 	fz_catch(ctx)
 	{
 		fz_drop_document(ctx, doc);
-		fprintf(stderr, "error: cannot draw '%s'\n", filename);
+		fz_log_error_printf(ctx, "cannot draw '%s'", filename);
 		errored = 1;
 	}
 
