@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2023 Artifex Software, Inc.
+// Copyright (C) 2004-2024 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -52,6 +52,7 @@ int pdftrim_main(int argc, char *argv[]);
 int pdfbake_main(int argc, char *argv[]);
 
 int cmapdump_main(int argc, char *argv[]);
+int pdfaudit_main(int argc, char *argv[]);
 
 static struct {
 	int (*func)(int argc, char *argv[]);
@@ -59,21 +60,21 @@ static struct {
 	char *desc;
 } tools[] = {
 #if FZ_ENABLE_PDF
-	{ pdfclean_main, "clean", "rewrite pdf file" },
+	{ pdfclean_main, "clean", "rewrite PDF file" },
 #endif
 	{ muconvert_main, "convert", "convert document" },
 #if FZ_ENABLE_PDF
-	{ pdfcreate_main, "create", "create pdf document" },
+	{ pdfcreate_main, "create", "create PDF document" },
 #endif
 	{ mudraw_main, "draw", "convert document" },
 	{ mutrace_main, "trace", "trace device calls" },
 #if FZ_ENABLE_PDF
 	{ pdfextract_main, "extract", "extract font and image resources" },
-	{ pdfinfo_main, "info", "show information about pdf resources" },
-	{ pdfmerge_main, "merge", "merge pages from multiple pdf sources into a new pdf" },
-	{ pdfpages_main, "pages", "show information about pdf pages" },
+	{ pdfinfo_main, "info", "show information about PDF resources" },
+	{ pdfmerge_main, "merge", "merge pages from multiple PDF sources into a new PDF" },
+	{ pdfpages_main, "pages", "show information about PDF pages" },
 	{ pdfposter_main, "poster", "split large page into many tiles" },
-	{ pdfrecolor_main, "recolor", "Change colorspace of pdf document" },
+	{ pdfrecolor_main, "recolor", "change colorspace of PDF document" },
 	{ pdfsign_main, "sign", "manipulate PDF digital signatures" },
 	{ pdftrim_main, "trim", "trim PDF page contents" },
 	{ pdfbake_main, "bake", "bake PDF form into static content" },
@@ -82,10 +83,11 @@ static struct {
 	{ murun_main, "run", "run javascript" },
 #endif
 #if FZ_ENABLE_PDF
-	{ pdfshow_main, "show", "show internal pdf objects" },
+	{ pdfshow_main, "show", "show internal PDF objects" },
 #ifndef NDEBUG
 	{ cmapdump_main, "cmapdump", "dump CMap resource as C source file" },
 #endif
+	{ pdfaudit_main, "audit", "produce usage stats from PDF files" },
 #endif
 };
 
