@@ -1,7 +1,7 @@
 //---------------------------------------------------------------------------------
 //
 //  Little Color Management System
-//  Copyright (c) 1998-2017 Marti Maria Saguer
+//  Copyright (c) 1998-2022 Marti Maria Saguer
 //
 // Permission is hereby granted, free of charge, to any person obtaining
 // a copy of this software and associated documentation files (the "Software"),
@@ -78,8 +78,8 @@
    Matrix-shaper based
    -------------------
 
-   This is implemented both with /CIEBasedABC or /CIEBasedDEF on dependig
-   of profile implementation. Since here there are no interpolation tables, I do
+   This is implemented both with /CIEBasedABC or /CIEBasedDEF depending on the
+   profile implementation. Since here there are no interpolation tables, I do
    the conversion directly to XYZ
 
 
@@ -295,21 +295,7 @@ cmsUInt8Number Word2Byte(cmsUInt16Number w)
 }
 
 
-// Convert to byte (using ICC2 notation)
-/*
-static
-cmsUInt8Number L2Byte(cmsUInt16Number w)
-{
-    int ww = w + 0x0080;
-
-    if (ww > 0xFFFF) return 0xFF;
-
-    return (cmsUInt8Number) ((cmsUInt16Number) (ww >> 8) & 0xFF);
-}
-*/
-
 // Write a cooked byte
-
 static
 void WriteByte(cmsContext ContextID, cmsIOHANDLER* m, cmsUInt8Number b)
 {
@@ -326,13 +312,13 @@ void WriteByte(cmsContext ContextID, cmsIOHANDLER* m, cmsUInt8Number b)
 // ----------------------------------------------------------------- PostScript generation
 
 
-// Removes offending Carriage returns
+// Removes offending carriage returns
+
 static
-char* RemoveCR(cmsContext ContextID, const char* txt)
+char* RemoveCR(const char* txt)
 {
     static char Buffer[2048];
     char* pt;
-    cmsUNUSED_PARAMETER(ContextID);
 
     strncpy(Buffer, txt, 2047);
     Buffer[2047] = 0;
@@ -364,8 +350,8 @@ void EmitHeader(cmsContext ContextID, cmsIOHANDLER* m, const char* Title, cmsHPR
     _cmsIOPrintf(ContextID, m, "%%!PS-Adobe-3.0\n");
     _cmsIOPrintf(ContextID, m, "%%\n");
     _cmsIOPrintf(ContextID, m, "%% %s\n", Title);
-    _cmsIOPrintf(ContextID, m, "%% Source: %s\n", RemoveCR(ContextID, DescASCII));
-    _cmsIOPrintf(ContextID, m, "%%         %s\n", RemoveCR(ContextID, CopyrightASCII));
+    _cmsIOPrintf(ContextID, m, "%% Source: %s\n", RemoveCR(DescASCII));
+    _cmsIOPrintf(ContextID, m, "%%         %s\n", RemoveCR(CopyrightASCII));
     _cmsIOPrintf(ContextID, m, "%% Created: %s", ctime(&timer)); // ctime appends a \n!!!
     _cmsIOPrintf(ContextID, m, "%%\n");
     _cmsIOPrintf(ContextID, m, "%%%%BeginResource\n");
@@ -425,21 +411,6 @@ void EmitIntent(cmsContext ContextID, cmsIOHANDLER* m, cmsUInt32Number Rendering
 //        = Yn*( L* / 116) / 7.787      if (L*) < 6 / 29
 //
 
-/*
-static
-void EmitL2Y(cmsIOHANDLER* m)
-{
-    _cmsIOPrintf(ContextID, m,
-            "{ "
-                "100 mul 16 add 116 div "               // (L * 100 + 16) / 116
-                 "dup 6 29 div ge "                     // >= 6 / 29 ?
-                 "{ dup dup mul mul } "                 // yes, ^3 and done
-                 "{ 4 29 div sub 108 841 div mul } "    // no, slope limiting
-            "ifelse } bind ");
-}
-*/
-
-
 // Lab -> XYZ, see the discussion above
 
 static
@@ -460,12 +431,28 @@ void EmitLab2XYZ(cmsContext ContextID, cmsIOHANDLER* m)
     _cmsIOPrintf(ContextID, m, "]\n");
 }
 
+static
+void EmitSafeGuardBegin(cmsContext ContextID, cmsIOHANDLER* m, const char* name)
+{
+    _cmsIOPrintf(ContextID, m, "%%LCMS2: Save previous definition of %s on the operand stack\n", name);
+    _cmsIOPrintf(ContextID, m, "currentdict /%s known { /%s load } { null } ifelse\n", name, name);
+}
 
+static
+void EmitSafeGuardEnd(cmsContext ContextID, cmsIOHANDLER* m, const char* name, int depth)
+{
+    _cmsIOPrintf(ContextID, m, "%%LCMS2: Restore previous definition of %s\n", name);
+    if (depth > 1) {
+        // cycle topmost items on the stack to bring the previous definition to the front
+        _cmsIOPrintf(ContextID, m, "%d -1 roll ", depth);
+    }
+    _cmsIOPrintf(ContextID, m, "dup null eq { pop currentdict /%s undef } { /%s exch def } ifelse\n", name, name);
+}
 
 // Outputs a table of words. It does use 16 bits
 
 static
-void Emit1Gamma(cmsContext ContextID, cmsIOHANDLER* m, cmsToneCurve* Table)
+void Emit1Gamma(cmsContext ContextID, cmsIOHANDLER* m, cmsToneCurve* Table, const char* name)
 {
     cmsUInt32Number i;
     cmsFloat64Number gamma;
@@ -480,28 +467,33 @@ void Emit1Gamma(cmsContext ContextID, cmsIOHANDLER* m, cmsToneCurve* Table)
     // Check if is really an exponential. If so, emit "exp"
     gamma = cmsEstimateGamma(ContextID, Table, 0.001);
      if (gamma > 0) {
-            _cmsIOPrintf(ContextID, m, "{ %g exp } bind ", gamma);
+            _cmsIOPrintf(ContextID, m, "/%s { %g exp } bind def\n", name, gamma);
             return;
      }
 
-    _cmsIOPrintf(ContextID, m, "{ ");
+    EmitSafeGuardBegin(ContextID, m, "lcms2gammatable");
+    _cmsIOPrintf(ContextID, m, "/lcms2gammatable [");
+
+    for (i=0; i < Table->nEntries; i++) {
+    if (i % 10 == 0)
+            _cmsIOPrintf(ContextID, m, "\n  ");
+        _cmsIOPrintf(ContextID, m, "%d ", Table->Table16[i]);
+    }
+
+    _cmsIOPrintf(ContextID, m, "] def\n");
+
+
+    // Emit interpolation code
+
+    // PostScript code                            Stack
+    // ===============                            ========================
+                                                  // v
+    _cmsIOPrintf(ContextID, m, "/%s {\n  ", name);
 
     // Bounds check
     EmitRangeCheck(ContextID, m);
 
-    // Emit intepolation code
-
-    // PostScript code                      Stack
-    // ===============                      ========================
-                                            // v
-    _cmsIOPrintf(ContextID, m, " [");
-
-    for (i=0; i < Table->nEntries; i++) {
-        _cmsIOPrintf(ContextID, m, "%d ", Table->Table16[i]);
-    }
-
-    _cmsIOPrintf(ContextID, m, "] ");                        // v tab
-
+    _cmsIOPrintf(ContextID, m, "\n  //lcms2gammatable ");    // v tab
     _cmsIOPrintf(ContextID, m, "dup ");                      // v tab tab
     _cmsIOPrintf(ContextID, m, "length 1 sub ");             // v tab dom
     _cmsIOPrintf(ContextID, m, "3 -1 roll ");                // tab dom v
@@ -513,7 +505,7 @@ void Emit1Gamma(cmsContext ContextID, cmsIOHANDLER* m, cmsToneCurve* Table)
     _cmsIOPrintf(ContextID, m, "ceiling cvi ");              // tab val2 cell0 cell1
     _cmsIOPrintf(ContextID, m, "3 index ");                  // tab val2 cell0 cell1 tab
     _cmsIOPrintf(ContextID, m, "exch ");                     // tab val2 cell0 tab cell1
-    _cmsIOPrintf(ContextID, m, "get ");                      // tab val2 cell0 y1
+    _cmsIOPrintf(ContextID, m, "get\n  ");                   // tab val2 cell0 y1
     _cmsIOPrintf(ContextID, m, "4 -1 roll ");                // val2 cell0 y1 tab
     _cmsIOPrintf(ContextID, m, "3 -1 roll ");                // val2 y1 tab cell0
     _cmsIOPrintf(ContextID, m, "get ");                      // val2 y1 y0
@@ -526,48 +518,50 @@ void Emit1Gamma(cmsContext ContextID, cmsIOHANDLER* m, cmsToneCurve* Table)
     _cmsIOPrintf(ContextID, m, "sub ");                      // y0 (y1-y0) rest
     _cmsIOPrintf(ContextID, m, "mul ");                      // y0 t1
     _cmsIOPrintf(ContextID, m, "add ");                      // y
-    _cmsIOPrintf(ContextID, m, "65535 div ");                // result
+    _cmsIOPrintf(ContextID, m, "65535 div\n");               // result
 
-    _cmsIOPrintf(ContextID, m, " } bind ");
+    _cmsIOPrintf(ContextID, m, "} bind def\n");
+
+    EmitSafeGuardEnd(ContextID, m, "lcms2gammatable", 1);
 }
 
 
 // Compare gamma table
 
 static
-cmsBool GammaTableEquals(cmsContext ContextID, cmsUInt16Number* g1, cmsUInt16Number* g2, cmsUInt32Number nEntries)
+cmsBool GammaTableEquals(cmsUInt16Number* g1, cmsUInt16Number* g2, cmsUInt32Number nG1, cmsUInt32Number nG2)
 {
-    cmsUNUSED_PARAMETER(ContextID);
-    return memcmp(g1, g2, nEntries* sizeof(cmsUInt16Number)) == 0;
+    if (nG1 != nG2) return FALSE;
+    return memcmp(g1, g2, nG1 * sizeof(cmsUInt16Number)) == 0;
 }
 
 
 // Does write a set of gamma curves
 
 static
-void EmitNGamma(cmsContext ContextID, cmsIOHANDLER* m, cmsUInt32Number n, cmsToneCurve* g[])
+void EmitNGamma(cmsContext ContextID, cmsIOHANDLER* m, cmsUInt32Number n, cmsToneCurve* g[], const char* nameprefix)
 {
     cmsUInt32Number i;
+    static char buffer[2048];
 
     for( i=0; i < n; i++ )
     {
         if (g[i] == NULL) return; // Error
 
-        if (i > 0 && GammaTableEquals(ContextID, g[i-1]->Table16, g[i]->Table16, g[i]->nEntries)) {
+        if (i > 0 && GammaTableEquals(g[i-1]->Table16, g[i]->Table16, g[i-1]->nEntries, g[i]->nEntries)) {
 
-            _cmsIOPrintf(ContextID, m, "dup ");
+            _cmsIOPrintf(ContextID, m, "/%s%d /%s%d load def\n", nameprefix, i, nameprefix, i-1);
         }
         else {
-            Emit1Gamma(ContextID, m, g[i]);
+            snprintf(buffer, sizeof(buffer), "%s%d", nameprefix, (int) i);
+            buffer[sizeof(buffer)-1] = '\0';
+            Emit1Gamma(ContextID, m, g[i], buffer);
         }
     }
 
 }
 
 
-
-
-
 // Following code dumps a LUT onto memory stream
 
 
@@ -584,7 +578,7 @@ void EmitNGamma(cmsContext ContextID, cmsIOHANDLER* m, cmsUInt32Number n, cmsTon
 //  component. -1 is used to mark beginning of whole block.
 
 static
-int OutputValueSampler(cmsContext ContextID, register const cmsUInt16Number In[], register cmsUInt16Number Out[], register void* Cargo)
+int OutputValueSampler(cmsContext ContextID, CMSREGISTER const cmsUInt16Number In[], CMSREGISTER cmsUInt16Number Out[], CMSREGISTER void* Cargo)
 {
     cmsPsSamplerCargo* sc = (cmsPsSamplerCargo*) Cargo;
     cmsUInt32Number i;
@@ -664,11 +658,11 @@ int OutputValueSampler(cmsContext ContextID, register const cmsUInt16Number In[]
 
 static
 void WriteCLUT(cmsContext ContextID, cmsIOHANDLER* m, cmsStage* mpe, const char* PreMaj,
-                                             const char* PostMaj,
-                                             const char* PreMin,
-                                             const char* PostMin,
-                                             int FixWhite,
-                                             cmsColorSpaceSignature ColorSpace)
+                                               const char* PostMaj,
+                                               const char* PreMin,
+                                               const char* PostMin,
+                                               int FixWhite,
+                                               cmsColorSpaceSignature ColorSpace)
 {
     cmsUInt32Number i;
     cmsPsSamplerCargo sc;
@@ -710,11 +704,11 @@ int EmitCIEBasedA(cmsContext ContextID, cmsIOHANDLER* m, cmsToneCurve* Curve, cm
     _cmsIOPrintf(ContextID, m, "[ /CIEBasedA\n");
     _cmsIOPrintf(ContextID, m, "  <<\n");
 
-    _cmsIOPrintf(ContextID, m, "/DecodeA ");
+    EmitSafeGuardBegin(ContextID, m, "lcms2gammaproc");
+    Emit1Gamma(ContextID, m, Curve, "lcms2gammaproc");
 
-    Emit1Gamma(ContextID, m, Curve);
-
-    _cmsIOPrintf(ContextID, m, " \n");
+    _cmsIOPrintf(ContextID, m, "/DecodeA /lcms2gammaproc load\n");
+    EmitSafeGuardEnd(ContextID, m, "lcms2gammaproc", 3);
 
     _cmsIOPrintf(ContextID, m, "/MatrixA [ 0.9642 1.0000 0.8249 ]\n");
     _cmsIOPrintf(ContextID, m, "/RangeLMN [ 0.0 0.9642 0.0 1.0000 0.0 0.8249 ]\n");
@@ -738,11 +732,19 @@ int EmitCIEBasedABC(cmsContext ContextID, cmsIOHANDLER* m, cmsFloat64Number* Mat
 
     _cmsIOPrintf(ContextID, m, "[ /CIEBasedABC\n");
     _cmsIOPrintf(ContextID, m, "<<\n");
-    _cmsIOPrintf(ContextID, m, "/DecodeABC [ ");
-
-    EmitNGamma(ContextID, m, 3, CurveSet);
 
+    EmitSafeGuardBegin(ContextID, m, "lcms2gammaproc0");
+    EmitSafeGuardBegin(ContextID, m, "lcms2gammaproc1");
+    EmitSafeGuardBegin(ContextID, m, "lcms2gammaproc2");
+    EmitNGamma(ContextID, m, 3, CurveSet, "lcms2gammaproc");
+    _cmsIOPrintf(ContextID, m, "/DecodeABC [\n");
+    _cmsIOPrintf(ContextID, m, "   /lcms2gammaproc0 load\n");
+    _cmsIOPrintf(ContextID, m, "   /lcms2gammaproc1 load\n");
+    _cmsIOPrintf(ContextID, m, "   /lcms2gammaproc2 load\n");
     _cmsIOPrintf(ContextID, m, "]\n");
+    EmitSafeGuardEnd(ContextID, m, "lcms2gammaproc2", 3);
+    EmitSafeGuardEnd(ContextID, m, "lcms2gammaproc1", 3);
+    EmitSafeGuardEnd(ContextID, m, "lcms2gammaproc0", 3);
 
     _cmsIOPrintf(ContextID, m, "/MatrixABC [ " );
 
@@ -774,28 +776,31 @@ int EmitCIEBasedDEF(cmsContext ContextID, cmsIOHANDLER* m, cmsPipeline* Pipeline
 {
     const char* PreMaj;
     const char* PostMaj;
-    const char* PreMin, *PostMin;
+    const char* PreMin, * PostMin;
     cmsStage* mpe;
+    int i, numchans;
+    static char buffer[2048];
 
-    mpe = Pipeline ->Elements;
+    mpe = Pipeline->Elements;
 
     switch (cmsStageInputChannels(ContextID, mpe)) {
     case 3:
+        _cmsIOPrintf(ContextID, m, "[ /CIEBasedDEF\n");
+        PreMaj = "<";
+        PostMaj = ">\n";
+        PreMin = PostMin = "";
+        break;
 
-            _cmsIOPrintf(ContextID, m, "[ /CIEBasedDEF\n");
-            PreMaj ="<";
-            PostMaj= ">\n";
-            PreMin = PostMin = "";
-            break;
     case 4:
-            _cmsIOPrintf(ContextID, m, "[ /CIEBasedDEFG\n");
-            PreMaj = "[";
-            PostMaj = "]\n";
-            PreMin = "<";
-            PostMin = ">\n";
-            break;
+        _cmsIOPrintf(ContextID, m, "[ /CIEBasedDEFG\n");
+        PreMaj = "[";
+        PostMaj = "]\n";
+        PreMin = "<";
+        PostMin = ">\n";
+        break;
+
     default:
-            return 0;
+        return 0;
 
     }
 
@@ -803,18 +808,34 @@ int EmitCIEBasedDEF(cmsContext ContextID, cmsIOHANDLER* m, cmsPipeline* Pipeline
 
     if (cmsStageType(ContextID, mpe) == cmsSigCurveSetElemType) {
 
-        _cmsIOPrintf(ContextID, m, "/DecodeDEF [ ");
-        EmitNGamma(ContextID, m, cmsStageOutputChannels(ContextID, mpe), _cmsStageGetPtrToCurveSet(mpe));
+        numchans = (int) cmsStageOutputChannels(ContextID, mpe);
+        for (i = 0; i < numchans; ++i) {
+            snprintf(buffer, sizeof(buffer), "lcms2gammaproc%d", i);
+            buffer[sizeof(buffer) - 1] = '\0';
+            EmitSafeGuardBegin(ContextID, m, buffer);
+        }
+        EmitNGamma(ContextID, m, cmsStageOutputChannels(ContextID, mpe), _cmsStageGetPtrToCurveSet(mpe), "lcms2gammaproc");
+        _cmsIOPrintf(ContextID, m, "/DecodeDEF [\n");
+        for (i = 0; i < numchans; ++i) {
+            snprintf(buffer, sizeof(buffer), "  /lcms2gammaproc%d load\n", i);
+            buffer[sizeof(buffer) - 1] = '\0';
+            _cmsIOPrintf(ContextID, m, buffer);
+        }
         _cmsIOPrintf(ContextID, m, "]\n");
+        for (i = numchans - 1; i >= 0; --i) {
+            snprintf(buffer, sizeof(buffer), "lcms2gammaproc%d", i);
+            buffer[sizeof(buffer) - 1] = '\0';
+            EmitSafeGuardEnd(ContextID, m, buffer, 3);
+        }
 
-        mpe = mpe ->Next;
+        mpe = mpe->Next;
     }
 
     if (cmsStageType(ContextID, mpe) == cmsSigCLutElemType) {
 
-            _cmsIOPrintf(ContextID, m, "/Table ");
-            WriteCLUT(ContextID, m, mpe, PreMaj, PostMaj, PreMin, PostMin, FALSE, (cmsColorSpaceSignature) 0);
-            _cmsIOPrintf(ContextID, m, "]\n");
+        _cmsIOPrintf(ContextID, m, "/Table ");
+        WriteCLUT(ContextID, m, mpe, PreMaj, PostMaj, PreMin, PostMin, FALSE, (cmsColorSpaceSignature)0);
+        _cmsIOPrintf(ContextID, m, "]\n");
     }
 
     EmitLab2XYZ(ContextID, m);
@@ -833,8 +854,8 @@ static
 cmsToneCurve* ExtractGray2Y(cmsContext ContextID, cmsHPROFILE hProfile, cmsUInt32Number Intent)
 {
     cmsToneCurve* Out = cmsBuildTabulatedToneCurve16(ContextID, 256, NULL);
-    cmsHPROFILE hXYZ  = cmsCreateXYZProfile();
-    cmsHTRANSFORM xform = cmsCreateTransformTHR(ContextID, hProfile, TYPE_GRAY_8, hXYZ, TYPE_XYZ_DBL, Intent, cmsFLAGS_NOOPTIMIZE);
+    cmsHPROFILE hXYZ  = cmsCreateXYZProfile(ContextID);
+    cmsHTRANSFORM xform = cmsCreateTransform(ContextID, hProfile, TYPE_GRAY_8, hXYZ, TYPE_XYZ_DBL, Intent, cmsFLAGS_NOOPTIMIZE);
     int i;
 
     if (Out != NULL && xform != NULL) {
@@ -880,12 +901,12 @@ int WriteInputLUT(cmsContext ContextID, cmsIOHANDLER* m, cmsHPROFILE hProfile, c
     cmsDetectBlackPoint(ContextID, &BlackPointAdaptedToD50, hProfile, Intent, 0);
 
     // Adjust output to Lab4
-    hLab = cmsCreateLab4ProfileTHR(ContextID, NULL);
+    hLab = cmsCreateLab4Profile(ContextID, NULL);
 
     Profiles[0] = hProfile;
     Profiles[1] = hLab;
 
-    xform = cmsCreateMultiprofileTransform(Profiles, 2,  InputFormat, TYPE_Lab_DBL, Intent, 0);
+    xform = cmsCreateMultiprofileTransform(ContextID, Profiles, 2,  InputFormat, TYPE_Lab_DBL, Intent, 0);
     cmsCloseProfile(ContextID, hLab);
 
     if (xform == NULL) {
@@ -911,7 +932,7 @@ int WriteInputLUT(cmsContext ContextID, cmsIOHANDLER* m, cmsHPROFILE hProfile, c
             cmsPipeline* DeviceLink;
             _cmsTRANSFORM* v = (_cmsTRANSFORM*) xform;
 
-            DeviceLink = cmsPipelineDup(ContextID, v ->Lut);
+            DeviceLink = cmsPipelineDup(ContextID, v ->core->Lut);
             if (DeviceLink == NULL) return 0;
 
             dwFlags |= cmsFLAGS_FORCE_CLUT;
@@ -925,7 +946,7 @@ int WriteInputLUT(cmsContext ContextID, cmsIOHANDLER* m, cmsHPROFILE hProfile, c
 
     default:
 
-        cmsSignalError(ContextID, cmsERROR_COLORSPACE_CHECK, "Only 3, 4 channels supported for CSA. This profile has %d channels.", nChannels);
+        cmsSignalError(ContextID, cmsERROR_COLORSPACE_CHECK, "Only 3, 4 channels are supported for CSA. This profile has %d channels.", nChannels);
         return 0;
     }
 
@@ -936,10 +957,9 @@ int WriteInputLUT(cmsContext ContextID, cmsIOHANDLER* m, cmsHPROFILE hProfile, c
 }
 
 static
-cmsFloat64Number* GetPtrToMatrix(cmsContext ContextID, const cmsStage* mpe)
+cmsFloat64Number* GetPtrToMatrix(const cmsStage* mpe)
 {
     _cmsStageMatrixData* Data = (_cmsStageMatrixData*) mpe ->Data;
-    cmsUNUSED_PARAMETER(ContextID);
 
     return Data -> Double;
 }
@@ -969,7 +989,7 @@ int WriteInputMatrixShaper(cmsContext ContextID, cmsIOHANDLER* m, cmsHPROFILE hP
             cmsMAT3 Mat;
             int i, j;
 
-            memmove(&Mat, GetPtrToMatrix(ContextID, Matrix), sizeof(Mat));
+            memmove(&Mat, GetPtrToMatrix(Matrix), sizeof(Mat));
 
             for (i = 0; i < 3; i++)
                 for (j = 0; j < 3; j++)
@@ -985,7 +1005,7 @@ int WriteInputMatrixShaper(cmsContext ContextID, cmsIOHANDLER* m, cmsHPROFILE hP
             return 0;
         }
 
-        return rc;
+    return rc;
 }
 
 
@@ -1002,8 +1022,8 @@ int WriteNamedColorCSA(cmsContext ContextID, cmsIOHANDLER* m, cmsHPROFILE hNamed
     char ColorName[cmsMAX_PATH];
     cmsNAMEDCOLORLIST* NamedColorList;
 
-    hLab  = cmsCreateLab4ProfileTHR(ContextID, NULL);
-    xform = cmsCreateTransform(hNamedColor, TYPE_NAMED_COLOR_INDEX, hLab, TYPE_Lab_DBL, Intent, 0);
+    hLab  = cmsCreateLab4Profile(ContextID, NULL);
+    xform = cmsCreateTransform(ContextID, hNamedColor, TYPE_NAMED_COLOR_INDEX, hLab, TYPE_Lab_DBL, Intent, 0);
     if (xform == NULL) return 0;
 
     NamedColorList = cmsGetNamedColorList(xform);
@@ -1242,8 +1262,6 @@ void EmitPQRStage(cmsContext ContextID, cmsIOHANDLER* m, cmsHPROFILE hProfile, i
                     "exch pop exch pop exch pop exch pop } bind\n]\n");
 
         }
-
-
 }
 
 
@@ -1273,7 +1291,7 @@ void EmitXYZ2Lab(cmsContext ContextID, cmsIOHANDLER* m)
 // Due to impedance mismatch between XYZ and almost all RGB and CMYK spaces
 // I choose to dump LUTS in Lab instead of XYZ. There is still a lot of wasted
 // space on 3D CLUT, but since space seems not to be a problem here, 33 points
-// would give a reasonable accurancy. Note also that CRD tables must operate in
+// would give a reasonable accuracy. Note also that CRD tables must operate in
 // 8 bits.
 
 static
@@ -1294,7 +1312,7 @@ int WriteOutputLUT(cmsContext ContextID, cmsIOHANDLER* m, cmsHPROFILE hProfile,
     cmsColorSpaceSignature ColorSpace;
 
 
-    hLab = cmsCreateLab4ProfileTHR(ContextID, NULL);
+    hLab = cmsCreateLab4Profile(ContextID, NULL);
     if (hLab == NULL) return 0;
 
     OutputFormat = cmsFormatterForColorspaceOfProfile(ContextID, hProfile, 2, FALSE);
@@ -1313,7 +1331,7 @@ int WriteOutputLUT(cmsContext ContextID, cmsIOHANDLER* m, cmsHPROFILE hProfile,
     Profiles[0] = hLab;
     Profiles[1] = hProfile;
 
-    xform = cmsCreateMultiprofileTransformTHR(ContextID,
+    xform = cmsCreateMultiprofileTransform(ContextID,
                                               Profiles, 2, TYPE_Lab_DBL,
                                               OutputFormat, RelativeEncodingIntent, 0);
     cmsCloseProfile(ContextID, hLab);
@@ -1326,7 +1344,7 @@ int WriteOutputLUT(cmsContext ContextID, cmsIOHANDLER* m, cmsHPROFILE hProfile,
 
     // Get a copy of the internal devicelink
     v = (_cmsTRANSFORM*) xform;
-    DeviceLink = cmsPipelineDup(ContextID, v ->Lut);
+    DeviceLink = cmsPipelineDup(ContextID, v ->core->Lut);
     if (DeviceLink == NULL) return 0;
 
 
@@ -1386,11 +1404,10 @@ int WriteOutputLUT(cmsContext ContextID, cmsIOHANDLER* m, cmsHPROFILE hProfile,
 
 // Builds a ASCII string containing colorant list in 0..1.0 range
 static
-void BuildColorantList(cmsContext ContextID, char *Colorant, cmsUInt32Number nColorant, cmsUInt16Number Out[])
+void BuildColorantList(char *Colorant, cmsUInt32Number nColorant, cmsUInt16Number Out[])
 {
     char Buff[32];
     cmsUInt32Number j;
-    cmsUNUSED_PARAMETER(ContextID);
 
     Colorant[0] = 0;
     if (nColorant > cmsMAXCHANNELS)
@@ -1418,7 +1435,7 @@ int WriteNamedColorCRD(cmsContext ContextID, cmsIOHANDLER* m, cmsHPROFILE hNamed
     cmsUInt32Number i, nColors, nColorant;
     cmsUInt32Number OutputFormat;
     char ColorName[cmsMAX_PATH];
-    char Colorant[128];
+    char Colorant[512];
     cmsNAMEDCOLORLIST* NamedColorList;
 
 
@@ -1426,7 +1443,7 @@ int WriteNamedColorCRD(cmsContext ContextID, cmsIOHANDLER* m, cmsHPROFILE hNamed
     nColorant    = T_CHANNELS(OutputFormat);
 
 
-    xform = cmsCreateTransform(hNamedColor, TYPE_NAMED_COLOR_INDEX, NULL, OutputFormat, Intent, dwFlags);
+    xform = cmsCreateTransform(ContextID, hNamedColor, TYPE_NAMED_COLOR_INDEX, NULL, OutputFormat, Intent, dwFlags);
     if (xform == NULL) return 0;
 
 
@@ -1451,7 +1468,7 @@ int WriteNamedColorCRD(cmsContext ContextID, cmsIOHANDLER* m, cmsHPROFILE hNamed
                 continue;
 
         cmsDoTransform(ContextID, xform, In, Out, 1);
-        BuildColorantList(ContextID, Colorant, nColorant, Out);
+        BuildColorantList(Colorant, nColorant, Out);
         _cmsIOPrintf(ContextID, m, "  (%s) [ %s ]\n", ColorName, Colorant);
     }
 
@@ -1513,8 +1530,6 @@ cmsUInt32Number  GenerateCRD(cmsContext ContextID,
 
     // Finally, return used byte count
     return dwBytesUsed;
-
-    cmsUNUSED_PARAMETER(ContextID);
 }
 
 
